# -*- coding: utf-8 -*-

import json
import time
import uuid
from typing import List, Dict, Tuple, Iterable

from requests import Response
from requests.structures import CaseInsensitiveDict

from TM1py.Exceptions.Exceptions import TM1pyRestException, TM1pyException
from TM1py.Objects.Process import Process
from TM1py.Objects.ProcessDebugBreakpoint import ProcessDebugBreakpoint
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url, require_admin
from TM1py.Utils.Utils import require_version


class ProcessService(ObjectService):
    """ Service to handle Object Updates for TI Processes
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def get(self, name_process: str, **kwargs) -> Process:
        """ Get a process from TM1 Server
    
        :param name_process:
        :return: Instance of the TM1py.Process
        """
        url = format_url(
            "/api/v1/Processes('{}')?$select=*,UIData,VariablesUIData,"
            "DataSource/dataSourceNameForServer,"
            "DataSource/dataSourceNameForClient,"
            "DataSource/asciiDecimalSeparator,"
            "DataSource/asciiDelimiterChar,"
            "DataSource/asciiDelimiterType,"
            "DataSource/asciiHeaderRecords,"
            "DataSource/asciiQuoteCharacter,"
            "DataSource/asciiThousandSeparator,"
            "DataSource/view,"
            "DataSource/query,"
            "DataSource/userName,"
            "DataSource/password,"
            "DataSource/usesUnicode,"
            "DataSource/subset", name_process)
        response = self._rest.GET(url, **kwargs)
        return Process.from_dict(response.json())

    def get_all(self, skip_control_processes: bool = False, **kwargs) -> List[Process]:
        """ Get a processes from TM1 Server
    
        :param skip_control_processes: bool, True to exclude processes that begin with "}" or "{"
        :return: List, instances of the TM1py.Process
        """
        model_process_filter = "&$filter=startswith(Name,'}') eq false and startswith(Name,'{') eq false"

        url = "/api/v1/Processes?$select=*,UIData,VariablesUIData," \
              "DataSource/dataSourceNameForServer," \
              "DataSource/dataSourceNameForClient," \
              "DataSource/asciiDecimalSeparator," \
              "DataSource/asciiDelimiterChar," \
              "DataSource/asciiDelimiterType," \
              "DataSource/asciiHeaderRecords," \
              "DataSource/asciiQuoteCharacter," \
              "DataSource/asciiThousandSeparator," \
              "DataSource/view," \
              "DataSource/query," \
              "DataSource/userName," \
              "DataSource/password," \
              "DataSource/usesUnicode," \
              "DataSource/subset{}".format(model_process_filter if skip_control_processes else "")

        response = self._rest.GET(url, **kwargs)
        response_as_dict = response.json()
        return [Process.from_dict(p) for p in response_as_dict['value']]

    def get_all_names(self, skip_control_processes: bool = False, **kwargs) -> List[str]:
        """ Get List with all process names from TM1 Server
        
        :param skip_control_processes: bool, True to exclude processes that begin with "}" or "{"
        :Returns:
            List of Strings
        """
        model_process_filter = "&$filter=startswith(Name,'}') eq false and startswith(Name,'{') eq false"
        url = "/api/v1/Processes?$select=Name{}".format(model_process_filter if skip_control_processes else "")

        response = self._rest.GET(url, **kwargs)
        processes = list(process['Name'] for process in response.json()['value'])
        return processes

    def search_string_in_code(self, search_string: str, skip_control_processes: bool = False, **kwargs) -> List[str]:
        """ Ask TM1 Server for list of process names that contain string anywhere in code tabs: Prolog,Metadata,Data,Epilog
        will not search DataSource, Parameters, Variables, or Attributes

        :param search_string: case insensitive string to search for
        :param skip_control_processes: bool, True to exclude processes that begin with "}" or "{"
        :Returns:
            List of strings
        """
        search_string = search_string.lower().replace(' ', '')
        model_process_filter = "and (startswith(Name,'}') eq false and startswith(Name,'{') eq false)"
        url = format_url("/api/v1/Processes?$select=Name&$filter="
                         "contains(tolower(replace(PrologProcedure, ' ', '')),'{}') "
                         "or contains(tolower(replace(MetadataProcedure, ' ', '')),'{}') "
                         "or contains(tolower(replace(DataProcedure, ' ', '')),'{}') "
                         "or contains(tolower(replace(EpilogProcedure, ' ', '')),'{}')",
                         search_string, search_string, search_string, search_string
                         )
        url += "{}".format(model_process_filter if skip_control_processes else "")
        response = self._rest.GET(url, **kwargs)
        processes = list(process['Name'] for process in response.json()['value'])
        return processes

    def search_string_in_name(self, name_startswith: str = None, name_contains: Iterable = None,
                              name_contains_operator: str = 'and', skip_control_processes: bool = False,
                              **kwargs) -> List[str]:
        """ Ask TM1 Server for list of process names that contain or start with string

        :param name_startswith: str, process name begins with (case insensitive)
        :param name_contains: iterable, found anywhere in name (case insensitive)
        :param name_contains_operator: 'and' or 'or'
        :param skip_control_processes: bool, True to exclude processes that begin with "}" or "{"
        """
        model_process_filter = "and (startswith(Name,'}') eq false and startswith(Name,'{') eq false)"
        name_contains_operator = name_contains_operator.strip().lower()
        if name_contains_operator not in ("and", "or"):
            raise ValueError("'name_contains_operator' must be either 'AND' or 'OR'")

        url = "/api/v1/Processes?$select=Name"
        name_filters = []

        if name_startswith:
            name_filters.append(format_url("startswith(toupper(Name),toupper('{}'))", name_startswith))

        if name_contains:
            if isinstance(name_contains, str):
                name_filters.append(format_url("contains(toupper(Name),toupper('{}'))", name_contains))

            elif isinstance(name_contains, Iterable):
                name_contains_filters = [format_url("contains(toupper(Name),toupper('{}'))", wildcard)
                                         for wildcard in name_contains]
                name_filters.append("({})".format(f" {name_contains_operator} ".join(name_contains_filters)))

            else:
                raise ValueError("'name_contains' must be str or iterable")

        url += "&$filter={}".format(f" and ".join(name_filters))
        url += "{}".format(model_process_filter if skip_control_processes else "")
        response = self._rest.GET(url, **kwargs)
        return list(process['Name'] for process in response.json()['value'])

    def create(self, process: Process, **kwargs) -> Response:
        """ Create a new process on TM1 Server

        :param process: Instance of TM1py.Process class
        :return: Response
        """
        url = "/api/v1/Processes"
        # Adjust process body if TM1 version is lower than 11 due to change in Process Parameters structure
        # https://www.ibm.com/developerworks/community/forums/html/topic?id=9188d139-8905-4895-9229-eaaf0e7fa683
        if int(self.version[0:2]) < 11:
            process.drop_parameter_types()
        response = self._rest.POST(url, process.body, **kwargs)
        return response

    def update(self, process: Process, **kwargs) -> Response:
        """ Update an existing Process on TM1 Server
    
        :param process: Instance of TM1py.Process class
        :return: Response
        """
        url = format_url("/api/v1/Processes('{}')", process.name)
        # Adjust process body if TM1 version is lower than 11 due to change in Process Parameters structure
        # https://www.ibm.com/developerworks/community/forums/html/topic?id=9188d139-8905-4895-9229-eaaf0e7fa683
        if int(self.version[0:2]) < 11:
            process.drop_parameter_types()
        response = self._rest.PATCH(url, process.body, **kwargs)
        return response

    def update_or_create(self, process: Process, **kwargs) -> Response:
        """ Update or Create a Process on TM1 Server

        :param process: Instance of TM1py.Process class
        :return: Response
        """
        if self.exists(name=process.name, **kwargs):
            return self.update(process=process, **kwargs)

        return self.create(process=process, **kwargs)

    def delete(self, name: str, **kwargs) -> Response:
        """ Delete a process in TM1
        
        :param name: 
        :return: Response
        """
        url = format_url("/api/v1/Processes('{}')", name)
        response = self._rest.DELETE(url, **kwargs)
        return response

    def exists(self, name: str, **kwargs) -> bool:
        """ Check if Process exists.
        
        :param name: 
        :return: 
        """
        url = format_url("/api/v1/Processes('{}')", name)
        return self._exists(url, **kwargs)

    def compile(self, name: str, **kwargs) -> List:
        """ Compile a Process. Return List of Syntax errors.
        
        :param name: 
        :return: 
        """
        url = format_url("/api/v1/Processes('{}')/tm1.Compile", name)
        response = self._rest.POST(url, **kwargs)
        syntax_errors = response.json()["value"]
        return syntax_errors

    def compile_process(self, process: Process, **kwargs) -> List:
        """ Compile a Process. Return List of Syntax errors.

        :param process:
        :return:
        """
        url = "/api/v1/CompileProcess"

        payload = json.loads('{"Process":' + process.body + '}')

        response = self._rest.POST(
            url=url,
            data=json.dumps(payload, ensure_ascii=False),
            **kwargs)

        syntax_errors = response.json()["value"]
        return syntax_errors

    def execute(self, process_name: str, parameters: Dict = None, timeout: float = None,
                cancel_at_timeout: bool = False, **kwargs) -> Response:
        """ Ask TM1 Server to execute a process. Call with parameter names as keyword arguments:
        tm1.processes.execute("Bedrock.Server.Wait", pLegalEntity="UK01")

        :param process_name:
        :param parameters: Deprecated! dictionary, e.g. {"Parameters": [ { "Name": "pLegalEntity", "Value": "UK01" }] }
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :param cancel_at_timeout: Abort operation in TM1 when timeout is reached
        :return:
        """
        url = format_url("/api/v1/Processes('{}')/tm1.Execute", process_name)
        if not parameters:
            if kwargs:
                parameters = {"Parameters": []}
                for parameter_name, parameter_value in kwargs.items():
                    parameters["Parameters"].append({"Name": parameter_name, "Value": parameter_value})
            else:
                parameters = {}
        return self._rest.POST(url=url, data=json.dumps(parameters, ensure_ascii=False), timeout=timeout,
                               cancel_at_timeout=cancel_at_timeout, **kwargs)

    @require_version(version="11.3")
    def execute_process_with_return(self, process: Process, timeout: float = None, cancel_at_timeout: bool = False,
                                    **kwargs) -> Tuple[bool, str, str]:
        """Run unbound TI code directly.

        :param process: a TI Process Object
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :param cancel_at_timeout: Abort operation in TM1 when timeout is reached
        :param kwargs: dictionary of process parameters and values
        :return: success (boolean), status (String), error_log_file (String)
        """
        url = "/api/v1/ExecuteProcessWithReturn?$expand=*"
        if kwargs:
            for parameter_name, parameter_value in kwargs.items():
                process.remove_parameter(name=parameter_name)
                process.add_parameter(name=parameter_name,
                                      prompt=parameter_name,
                                      value=parameter_value)

        payload = json.loads("{\"Process\":" + process.body + "}")

        response = self._rest.POST(
            url=url,
            data=json.dumps(payload, ensure_ascii=False),
            timeout=timeout,
            cancel_at_timeout=cancel_at_timeout,
            **kwargs)

        execution_summary = response.json()

        success = execution_summary["ProcessExecuteStatusCode"] == "CompletedSuccessfully"
        status = execution_summary["ProcessExecuteStatusCode"]
        error_log_file = None if execution_summary["ErrorLogFile"] is None else execution_summary["ErrorLogFile"][
            "Filename"]
        return success, status, error_log_file

    def execute_with_return(self, process_name: str, timeout: float = None, cancel_at_timeout: bool = False,
                            **kwargs) -> Tuple[bool, str, str]:
        """ Ask TM1 Server to execute a process.
        pass process parameters as keyword arguments to this function. E.g:

        self.tm1.processes.execute_with_return(
            process_name="Bedrock.Server.Wait",
            pWaitSec=2)

        :param process_name: name of the TI process
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :param cancel_at_timeout: Abort operation in TM1 when timeout is reached
        :param kwargs: dictionary of process parameters and values
        :return: success (boolean), status (String), error_log_file (String)
        """
        url = format_url("/api/v1/Processes('{}')/tm1.ExecuteWithReturn?$expand=*", process_name)
        parameters = dict()
        if kwargs:
            parameters = {"Parameters": []}
            for parameter_name, parameter_value in kwargs.items():
                parameters["Parameters"].append({"Name": parameter_name, "Value": parameter_value})

        response = self._rest.POST(
            url=url,
            data=json.dumps(parameters, ensure_ascii=False),
            timeout=timeout,
            cancel_at_timeout=cancel_at_timeout,
            **kwargs)
        execution_summary = response.json()
        success = execution_summary["ProcessExecuteStatusCode"] == "CompletedSuccessfully"
        status = execution_summary["ProcessExecuteStatusCode"]
        error_log_file = None if execution_summary["ErrorLogFile"] is None else execution_summary["ErrorLogFile"][
            "Filename"]
        return success, status, error_log_file

    @require_admin
    def execute_ti_code(self, lines_prolog: Iterable[str], lines_epilog: Iterable[str] = None, **kwargs) -> Response:
        """ Execute lines of code on the TM1 Server

            :param lines_prolog: list - where each element is a valid statement of TI code.
            :param lines_epilog: list - where each element is a valid statement of TI code.
        """
        process_name = "".join(['}TM1py', str(uuid.uuid4())])
        p = Process(name=process_name,
                    prolog_procedure=Process.AUTO_GENERATED_STATEMENTS + '\r\n'.join(lines_prolog),
                    epilog_procedure=Process.AUTO_GENERATED_STATEMENTS + '\r\n'.join(
                        lines_epilog) if lines_epilog else '')
        self.create(p, **kwargs)
        try:
            return self.execute(process_name, **kwargs)
        except TM1pyRestException as e:
            raise e
        finally:
            self.delete(process_name, **kwargs)

    def get_error_log_file_content(self, file_name: str, **kwargs) -> str:
        """ Get content of error log file (e.g. TM1ProcessError_20180926213819_65708356_979b248b-232e622c6.log)

        :param file_name: name of the error log file in the TM1 log directory
        :return: String, content of the file
        """
        url = format_url("/api/v1/ErrorLogFiles('{}')/Content", file_name)
        response = self._rest.GET(url=url, **kwargs)
        return response.text

    def get_processerrorlogs(self, process_name: str, **kwargs) -> List:
        """ Get all ProcessErrorLog entries for a process

        :param process_name: name of the process
        :return: list - Collection of ProcessErrorLogs
        """
        url = format_url("/api/v1/Processes('{}')/ErrorLogs", process_name)
        response = self._rest.GET(url=url, **kwargs)
        return response.json()['value']

    def get_last_message_from_processerrorlog(self, process_name: str, **kwargs) -> str:
        """ Get the latest ProcessErrorLog from a process entity

            :param process_name: name of the process
            :return: String - the errorlog, e.g.:  "Error: Data procedure line (9): Invalid key:
            Dimension Name: "Product", Element Name (Key): "ProductA""
        """
        logs_as_list = self.get_processerrorlogs(process_name, **kwargs)
        if len(logs_as_list) > 0:
            timestamp = logs_as_list[-1]['Timestamp']
            url = format_url("/api/v1/Processes('{}')/ErrorLogs('{}')/Content", process_name, timestamp)
            # response is plain text - due to entity type Edm.Stream
            response = self._rest.GET(url=url, **kwargs)
            return response.text

    def debug_process(self, process_name: str, timeout: float = None, **kwargs) -> Dict:
        """ 
        Start debug session for specified process; debug session id is returned in response        
        """
        raw_url = "/api/v1/Processes('{}')/tm1.Debug?$expand=Breakpoints," \
                  "Thread,CallStack($expand=Variables,Process($select=Name))"
        url = format_url(raw_url, process_name)

        parameters = dict()
        if kwargs:
            parameters = {"Parameters": []}
            for parameter_name, parameter_value in kwargs.items():
                parameters["Parameters"].append({"Name": parameter_name, "Value": parameter_value})

        response = self._rest.POST(
            url,
            data=json.dumps(parameters, ensure_ascii=False),
            timeout=timeout,
            **kwargs)
        return response.json()

    def debug_step_over(self, debug_id: str, **kwargs) -> Dict:
        """ 
        Runs a single statement in the process
        If ExecuteProcess is next function, will NOT debug child process
        """
        url = format_url("/api/v1/ProcessDebugContexts('{}')/tm1.StepOver", debug_id)
        self._rest.POST(url, **kwargs)

        # digest time  necessary for TM1 <= 11.8
        # ToDo: remove in later versions of TM1 once issue in TM1 server is resolved
        time.sleep(0.1)

        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=Breakpoints," \
                  "Thread,CallStack($expand=Variables,Process($select=Name))"
        url = format_url(raw_url, debug_id)
        response = self._rest.GET(url, **kwargs)

        return response.json()

    def debug_step_in(self, debug_id: str, **kwargs) -> Dict:
        """ 
        Runs a single statement in the process
        If ExecuteProcess is next function, will pause at first statement inside child process
        """
        url = format_url("/api/v1/ProcessDebugContexts('{}')/tm1.StepIn", debug_id)
        self._rest.POST(url, **kwargs)

        # digest time  necessary for TM1 <= 11.8
        # ToDo: remove in later versions of TM1 once issue in TM1 server is resolved
        time.sleep(0.1)

        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=Breakpoints," \
                  "Thread,CallStack($expand=Variables,Process($select=Name))"
        url = format_url(raw_url, debug_id)
        response = self._rest.GET(url, **kwargs)

        return response.json()

    def debug_step_out(self, debug_id: str, **kwargs) -> Dict:
        """
        Resumes execution and runs until current process has finished.
        """
        url = format_url("/api/v1/ProcessDebugContexts('{}')/tm1.StepOut", debug_id)
        self._rest.POST(url, **kwargs)

        # digest time  necessary for TM1 <= 11.8
        # ToDo: remove in later versions of TM1 once issue in TM1 server is resolved
        time.sleep(0.1)

        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=Breakpoints," \
                  "Thread,CallStack($expand=Variables,Process($select=Name))"
        url = format_url(raw_url, debug_id)
        response = self._rest.GET(url, **kwargs)

        return response.json()

    def debug_continue(self, debug_id: str, **kwargs) -> Dict:
        """ 
        Resumes execution until next breakpoint
        
        """
        url = format_url("/api/v1/ProcessDebugContexts('{}')/tm1.Continue", debug_id)
        self._rest.POST(url, **kwargs)

        # digest time  necessary for TM1 <= 11.8
        # ToDo: remove in later versions of TM1 once issue in TM1 server is resolved
        time.sleep(0.1)

        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=Breakpoints," \
                  "Thread,CallStack($expand=Variables,Process($select=Name))"
        url = format_url(raw_url, debug_id)
        response = self._rest.GET(url, **kwargs)

        return response.json()

    def debug_get_breakpoints(self, debug_id: str, **kwargs) -> List[ProcessDebugBreakpoint]:
        url = format_url("/api/v1/ProcessDebugContexts('{}')/Breakpoints", debug_id)

        response = self._rest.GET(url, **kwargs)
        return [ProcessDebugBreakpoint.from_dict(b) for b in response.json()['value']]

    def debug_add_breakpoint(self, debug_id: str, break_point: ProcessDebugBreakpoint, **kwargs) -> Response:
        return self.debug_add_breakpoints(debug_id, [break_point], **kwargs)

    def debug_add_breakpoints(self, debug_id: str, break_points: Iterable[ProcessDebugBreakpoint] = None,
                              **kwargs) -> Response:
        url = format_url("/api/v1/ProcessDebugContexts('{}')/Breakpoints", debug_id)

        body = json.dumps([break_point.body_as_dict for break_point in break_points])

        response = self._rest.POST(url, body, **kwargs)
        return response

    def debug_remove_breakpoint(self, debug_id: str, breakpoint_id: int, **kwargs) -> Response:
        url = format_url("/api/v1/ProcessDebugContexts('{}')/Breakpoints('{}')", debug_id, str(breakpoint_id))

        response = self._rest.DELETE(url, **kwargs)
        return response

    def debug_update_breakpoint(self, debug_id: str, break_point: ProcessDebugBreakpoint, **kwargs) -> Response:
        url = format_url(
            "/api/v1/ProcessDebugContexts('{}')/Breakpoints('{}')",
            debug_id,
            str(break_point.breakpoint_id))

        response = self._rest.PATCH(url, break_point.body, **kwargs)
        return response

    def debug_get_variable_values(self, debug_id: str, **kwargs) -> CaseInsensitiveDict:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=" \
                  "CallStack($expand=Variables)"
        url = format_url(raw_url, debug_id)

        response = self._rest.GET(url, **kwargs)
        result = response.json()
        call_stack = result['CallStack'][0]['Variables'] if result['CallStack'] else result['CallStack']

        return CaseInsensitiveDict({entry["Name"]: entry["Value"] for entry in call_stack})

    def debug_get_single_variable_value(self, debug_id: str, variable_name: str, **kwargs) -> str:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=" \
                  "CallStack($expand=Variables($filter=tolower(Name) eq '{}';$select=Value))"
        url = format_url(raw_url, debug_id, variable_name.lower())

        response = self._rest.GET(url, **kwargs)

        try:
            return response.json()['CallStack'][0]['Variables'][0]['Value']
        except:
            raise ValueError(f"'{variable_name}' not found in collection")

    def debug_get_process_procedure(self, debug_id: str, **kwargs) -> str:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=" \
                  "CallStack($select=Procedure)"
        url = format_url(raw_url, debug_id)

        response = self._rest.GET(url, **kwargs)
        return response.json()['CallStack'][0]['Procedure']

    def debug_get_process_line_number(self, debug_id: str, **kwargs) -> str:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=" \
                  "CallStack($select=LineNumber)"
        url = format_url(raw_url, debug_id)

        response = self._rest.GET(url, **kwargs)
        return response.json()['CallStack'][0]['LineNumber']

    def debug_get_record_number(self, debug_id: str, **kwargs) -> str:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=" \
                  "CallStack($select=RecordNumber)"
        url = format_url(raw_url, debug_id)

        response = self._rest.GET(url, **kwargs)
        return response.json()['CallStack'][0]['RecordNumber']

    def debug_get_current_breakpoint(self, debug_id: str, **kwargs) -> ProcessDebugBreakpoint:
        raw_url = "/api/v1/ProcessDebugContexts('{}')?$expand=CurrentBreakpoint"

        url = format_url(raw_url, debug_id)

        response = self._rest.GET(url=url, **kwargs)
        return ProcessDebugBreakpoint.from_dict(response.json()["CurrentBreakpoint"])

    @require_admin
    def evaluate_boolean_ti_expression(self, formula: str):
        prolog_procedure = f"""
        if (~{formula.strip(";")});
          ProcessQuit;
        endif;
        """

        process = Process(name="", prolog_procedure=prolog_procedure)
        success, status, _ = self.execute_process_with_return(process)
        if status == "QuitCalled":
            return False
        elif status == "CompletedSuccessfully":
            return True
        else:
            raise TM1pyException(f"Unexpected TI return status: '{status}' for expression: '{formula}'")

    @require_admin
    def evaluate_ti_expression(self, formula: str, **kwargs) -> str:
        """ This function is same functionality as hitting "Evaluate" within variable formula editor in TI
            Function creates temporary TI and then starts a debug session on that TI
            EnableTIDebugging=T must be present in .cfg file
            Only suited for DEV and one-off uses, don't incorporate into dataframe lambda function

        :param formula: a valid tm1 variable formula (no double quotes, no equals sign, semicolon optional)
            e.g. "8*2;", "CellGetN('c1', 'e1', 'e2);", "ATTRS('Region', 'France', 'Currency')"
        :returns: string result from formula
        """

        # grab everything to right of "=" if present
        formula = formula[formula.find('=') + 1:]

        # make sure semicolon at end is present
        if not formula.strip().endswith(";"):
            formula += ";"

        prolog_list = ["sFunc = {}".format(formula), "sDebug='Stop';"]
        process_name = "".join(['}TM1py', str(uuid.uuid4())])
        p = Process(
            name=process_name,
            prolog_procedure=Process.AUTO_GENERATED_STATEMENTS + '\r\n'.join(prolog_list))
        syntax_errors = self.compile_process(p, **kwargs)

        if syntax_errors:
            raise ValueError(str(syntax_errors))

        try:
            self.create(p, **kwargs)
            debug_id = self.debug_process(p.name, **kwargs)['ID']
            break_point = ProcessDebugBreakpoint(
                breakpoint_id=1,
                breakpoint_type='ProcessDebugContextDataBreakpoint',
                enabled=True,
                hit_mode='BreakAlways',
                variable_name='sFunc')
            self.debug_add_breakpoint(debug_id=debug_id, break_point=break_point, **kwargs)
            self.debug_continue(debug_id, **kwargs)
            result = self.debug_get_variable_values(debug_id, **kwargs)
            self.debug_continue(debug_id, **kwargs)

            if not result:
                raise ValueError('unknown error: no formula result found')
            else:
                return result['sFunc']

        except TM1pyRestException as e:
            raise e

        finally:
            self.delete(p.name, **kwargs)



# -*- coding: utf-8 -*-

import json
import re
from typing import Optional, Iterable, Dict, List, Union

from TM1py.Objects.TM1Object import TM1Object


class Process(TM1Object):
    """ Abstraction of a TM1 Process.

        IMPORTANT. doesn't work with Processes that were generated through the Wizard
    """

    """ the auto_generated_string code is required to be in all code-tabs. """
    BEGIN_GENERATED_STATEMENTS = "#****Begin: Generated Statements***"
    END_GENERATED_STATEMENTS = "#****End: Generated Statements****"
    AUTO_GENERATED_STATEMENTS = "{}\r\n{}\r\n".format(BEGIN_GENERATED_STATEMENTS, END_GENERATED_STATEMENTS)
    MAX_STATEMENTS = 16_380
    MAX_STATEMENTS_POST_11_8_15 = 100_000

    @staticmethod
    def add_generated_string_to_code(code: str) -> str:
        pattern = r"(?s)#\*\*\*\*Begin: Generated Statements(.*)#\*\*\*\*End: Generated Statements\*\*\*\*"
        if re.search(pattern=pattern, string=code):
            return code
        else:
            return Process.AUTO_GENERATED_STATEMENTS + code

    def __init__(self,
                 name: str,
                 has_security_access: Optional[bool] = False,
                 ui_data: str = "CubeAction=1511\fDataAction=1503\fCubeLogChanges=0\f",
                 parameters: Iterable = None,
                 variables: Iterable = None,
                 variables_ui_data: Iterable = None,
                 prolog_procedure: str = '',
                 metadata_procedure: str = '',
                 data_procedure: str = '',
                 epilog_procedure: str = '',
                 datasource_type: str = 'None',
                 datasource_ascii_decimal_separator: str = '.',
                 datasource_ascii_delimiter_char: str = ';',
                 datasource_ascii_delimiter_type: str = 'Character',
                 datasource_ascii_header_records: int = 1,
                 datasource_ascii_quote_character: str = '',
                 datasource_ascii_thousand_separator: str = ',',
                 datasource_data_source_name_for_client: str = '',
                 datasource_data_source_name_for_server: str = '',
                 datasource_password: str = '',
                 datasource_user_name: str = '',
                 datasource_query: str = '',
                 datasource_uses_unicode: bool = True,
                 datasource_view: str = '',
                 datasource_subset: str = ''):
        """ Default construcor

        :param name: name of the process - mandatory
        :param has_security_access:
        :param ui_data:
        :param parameters:
        :param variables:
        :param variables_ui_data:
        :param prolog_procedure:
        :param metadata_procedure:
        :param data_procedure:
        :param epilog_procedure:
        :param datasource_type:
        :param datasource_ascii_decimal_separator:
        :param datasource_ascii_delimiter_char:
        :param datasource_ascii_delimiter_type:
        :param datasource_ascii_header_records:
        :param datasource_ascii_quote_character:
        :param datasource_ascii_thousand_separator:
        :param datasource_data_source_name_for_client:
        :param datasource_data_source_name_for_server:
        :param datasource_password:
        :param datasource_user_name:
        :param datasource_query:
        :param datasource_uses_unicode:
        :param datasource_view:
        :param datasource_subset:
        """
        self._name = name
        self._has_security_access = has_security_access
        self._ui_data = ui_data
        self._parameters = list(parameters) if parameters else []
        self._variables = list(variables) if variables else []
        if variables_ui_data:
            # Handle encoding issue in variable_ui_data for async requests
            self._variables_ui_data = [entry.replace("€", "\f") for entry in variables_ui_data]
        else:
            self._variables_ui_data = []
        self._prolog_procedure = Process.add_generated_string_to_code(prolog_procedure)
        self._metadata_procedure = Process.add_generated_string_to_code(metadata_procedure)
        self._data_procedure = Process.add_generated_string_to_code(data_procedure)
        self._epilog_procedure = Process.add_generated_string_to_code(epilog_procedure)
        self._datasource_type = datasource_type
        self._datasource_ascii_decimal_separator = datasource_ascii_decimal_separator
        self._datasource_ascii_delimiter_char = datasource_ascii_delimiter_char
        self._datasource_ascii_delimiter_type = datasource_ascii_delimiter_type
        self._datasource_ascii_header_records = datasource_ascii_header_records
        self._datasource_ascii_quote_character = datasource_ascii_quote_character
        self._datasource_ascii_thousand_separator = datasource_ascii_thousand_separator
        self._datasource_data_source_name_for_client = datasource_data_source_name_for_client
        self._datasource_data_source_name_for_server = datasource_data_source_name_for_server
        self._datasource_password = datasource_password
        self._datasource_user_name = datasource_user_name
        self._datasource_query = datasource_query
        self._datasource_uses_unicode = datasource_uses_unicode
        self._datasource_view = datasource_view
        self._datasource_subset = datasource_subset

    @classmethod
    def from_json(cls, process_as_json: str) -> 'Process':
        """
        :param process_as_json: response of /api/v1/Processes('x')?$expand=*
        :return: an instance of this class
        """
        process_as_dict = json.loads(process_as_json)
        return cls.from_dict(process_as_dict)

    @classmethod
    def from_dict(cls, process_as_dict: Dict) -> 'Process':
        """
        :param process_as_dict: Dictionary, process as dictionary
        :return: an instance of this class
        """
        return cls(name=process_as_dict['Name'],
                   has_security_access=process_as_dict['HasSecurityAccess'],
                   ui_data=process_as_dict['UIData'],
                   parameters=process_as_dict['Parameters'],
                   variables=process_as_dict['Variables'],
                   variables_ui_data=process_as_dict['VariablesUIData'],
                   prolog_procedure=process_as_dict['PrologProcedure'],
                   metadata_procedure=process_as_dict['MetadataProcedure'],
                   data_procedure=process_as_dict['DataProcedure'],
                   epilog_procedure=process_as_dict['EpilogProcedure'],
                   datasource_type=process_as_dict['DataSource'].get('Type', ''),
                   datasource_ascii_decimal_separator=process_as_dict['DataSource'].get('asciiDecimalSeparator', ''),
                   datasource_ascii_delimiter_char=process_as_dict['DataSource'].get('asciiDelimiterChar', ''),
                   datasource_ascii_delimiter_type=process_as_dict['DataSource'].get('asciiDelimiterType', ''),
                   datasource_ascii_header_records=process_as_dict['DataSource'].get('asciiHeaderRecords', ''),
                   datasource_ascii_quote_character=process_as_dict['DataSource'].get('asciiQuoteCharacter', ''),
                   datasource_ascii_thousand_separator=process_as_dict['DataSource'].get('asciiThousandSeparator', ''),
                   datasource_data_source_name_for_client=process_as_dict['DataSource'].get('dataSourceNameForClient',''),
                   datasource_data_source_name_for_server=process_as_dict['DataSource'].get('dataSourceNameForServer',''),
                   datasource_password=process_as_dict['DataSource'].get('password', ''),
                   datasource_user_name=process_as_dict['DataSource'].get('userName', ''),
                   datasource_query=process_as_dict['DataSource'].get('query', ''),
                   datasource_uses_unicode=process_as_dict['DataSource'].get('usesUnicode', ''),
                   datasource_view=process_as_dict['DataSource'].get('view', ''),
                   datasource_subset=process_as_dict['DataSource'].get('subset', ''))

    @property
    def body(self) -> str:
        return self._construct_body()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def has_security_access(self) -> bool:
        return self._has_security_access

    @has_security_access.setter
    def has_security_access(self, value: bool):
        self._has_security_access = value

    @property
    def variables(self) -> List:
        return self._variables

    @property
    def parameters(self) -> List:
        return self._parameters

    @property
    def prolog_procedure(self) -> str:
        return self._prolog_procedure

    @prolog_procedure.setter
    def prolog_procedure(self, value: str):
        self._prolog_procedure = Process.add_generated_string_to_code(value)

    @property
    def metadata_procedure(self) -> str:
        return self._metadata_procedure

    @metadata_procedure.setter
    def metadata_procedure(self, value: str):
        self._metadata_procedure = Process.add_generated_string_to_code(value)

    @property
    def data_procedure(self) -> str:
        return self._data_procedure

    @data_procedure.setter
    def data_procedure(self, value: str):
        self._data_procedure = Process.add_generated_string_to_code(value)

    @property
    def epilog_procedure(self) -> str:
        return self._epilog_procedure

    @epilog_procedure.setter
    def epilog_procedure(self, value: str):
        self._epilog_procedure = Process.add_generated_string_to_code(value)

    @property
    def datasource_type(self) -> str:
        return self._datasource_type

    @datasource_type.setter
    def datasource_type(self, value: str):
        self._datasource_type = value

    @property
    def datasource_ascii_decimal_separator(self) -> str:
        return self._datasource_ascii_decimal_separator

    @datasource_ascii_decimal_separator.setter
    def datasource_ascii_decimal_separator(self, value: str):
        self._datasource_ascii_decimal_separator = value

    @property
    def datasource_ascii_delimiter_char(self) -> str:
        return self._datasource_ascii_delimiter_char

    @datasource_ascii_delimiter_char.setter
    def datasource_ascii_delimiter_char(self, value: str):
        self._datasource_ascii_delimiter_char = value

    @property
    def datasource_ascii_delimiter_type(self) -> str:
        return self._datasource_ascii_delimiter_type

    @datasource_ascii_delimiter_type.setter
    def datasource_ascii_delimiter_type(self, value: str):
        self._datasource_ascii_delimiter_type = value

    @property
    def datasource_ascii_header_records(self) -> int:
        return self._datasource_ascii_header_records

    @datasource_ascii_header_records.setter
    def datasource_ascii_header_records(self, value: int):
        self._datasource_ascii_header_records = value

    @property
    def datasource_ascii_quote_character(self) -> str:
        return self._datasource_ascii_quote_character

    @datasource_ascii_quote_character.setter
    def datasource_ascii_quote_character(self, value: str):
        self._datasource_ascii_quote_character = value

    @property
    def datasource_ascii_thousand_separator(self) -> str:
        return self._datasource_ascii_thousand_separator

    @datasource_ascii_thousand_separator.setter
    def datasource_ascii_thousand_separator(self, value: str):
        self._datasource_ascii_thousand_separator = value

    @property
    def datasource_data_source_name_for_client(self) -> str:
        return self._datasource_data_source_name_for_client

    @datasource_data_source_name_for_client.setter
    def datasource_data_source_name_for_client(self, value: str):
        self._datasource_data_source_name_for_client = value

    @property
    def datasource_data_source_name_for_server(self) -> str:
        return self._datasource_data_source_name_for_server

    @datasource_data_source_name_for_server.setter
    def datasource_data_source_name_for_server(self, value: str):
        self._datasource_data_source_name_for_server = value

    @property
    def datasource_password(self) -> str:
        return self._datasource_password

    @datasource_password.setter
    def datasource_password(self, value: str):
        self._datasource_password = value

    @property
    def datasource_user_name(self) -> str:
        return self._datasource_user_name

    @datasource_user_name.setter
    def datasource_user_name(self, value: str):
        self._datasource_user_name = value

    @property
    def datasource_query(self) -> str:
        return self._datasource_query

    @datasource_query.setter
    def datasource_query(self, value: str):
        self._datasource_query = value

    @property
    def datasource_uses_unicode(self) -> bool:
        return self._datasource_uses_unicode

    @datasource_uses_unicode.setter
    def datasource_uses_unicode(self, value: bool):
        self._datasource_uses_unicode = value

    @property
    def datasource_view(self) -> str:
        return self._datasource_view

    @datasource_view.setter
    def datasource_view(self, value: str):
        self._datasource_view = value

    @property
    def datasource_subset(self) -> str:
        return self._datasource_subset

    @datasource_subset.setter
    def datasource_subset(self, value: str):
        self._datasource_subset = value

    def add_variable(self, name: str, variable_type: str):
        """ add variable to the process

        :param name: -
        :param variable_type: 'String' or 'Numeric'
        :return:
        """
        # variable consists of actual variable and UI-Information ('ignore','other', etc.)
        # 1. handle Variable info
        variable = {'Name': name,
                    'Type': variable_type,
                    'Position': len(self._variables) + 1,
                    'StartByte': 0,
                    'EndByte': 0}
        self._variables.append(variable)
        # 2. handle UI info
        var_type = 33 if variable_type == 'Numeric' else 32
        # '\f' !
        variable_ui_data = 'VarType=' + str(var_type) + '\f' + 'ColType=' + str(827) + '\f'
        """
        mapping VariableUIData:
            VarType 33 -> Numeric
            VarType 32 -> String
            ColType 827 -> Other
        """
        self._variables_ui_data.append(variable_ui_data)

    def remove_variable(self, name: str):
        for variable in self.variables[:]:
            if variable['Name'] == name:
                vuid = self._variables_ui_data[self._variables.index(variable)]
                self._variables_ui_data.remove(vuid)
                self._variables.remove(variable)

    def add_parameter(self, name: str, prompt: str, value: Union[str, int, float],
                      parameter_type: Optional[str] = None):
        """
        
        :param name: 
        :param prompt: 
        :param value: 
        :param parameter_type: introduced in TM1 11 REST API, therefor optional. if Not given type is derived from value
        :return: 
        """
        if not parameter_type:
            parameter_type = 'String' if isinstance(value, str) else 'Numeric'
        parameter = {'Name': name,
                     'Prompt': prompt,
                     'Value': value,
                     'Type': parameter_type}
        self._parameters.append(parameter)

    def remove_parameter(self, name: str):
        for parameter in self.parameters[:]:
            if parameter['Name'] == name:
                self._parameters.remove(parameter)

    def drop_parameter_types(self):
        for p in range(len(self.parameters)):
            if 'Type' in self.parameters[p]:
                del self.parameters[p]['Type']

    # construct self.body (json) from the class-attributes
    def _construct_body(self) -> str:
        # general parameters
        body_as_dict = {
            'Name': self._name,
            'PrologProcedure': self._prolog_procedure,
            'MetadataProcedure': self._metadata_procedure,
            'DataProcedure': self._data_procedure,
            'EpilogProcedure': self._epilog_procedure,
            'HasSecurityAccess': self._has_security_access,
            'UIData': self._ui_data,
            'DataSource': {},
            'Parameters': self._parameters,
            'Variables': self._variables,
            'VariablesUIData': self._variables_ui_data}

        # specific parameters (depending on datasource type)
        if self._datasource_type == 'ASCII':
            body_as_dict['DataSource'] = {
                "Type": self._datasource_type,
                "asciiDecimalSeparator": self._datasource_ascii_decimal_separator,
                "asciiDelimiterChar": self._datasource_ascii_delimiter_char,
                "asciiDelimiterType": self._datasource_ascii_delimiter_type,
                "asciiHeaderRecords": self._datasource_ascii_header_records,
                "asciiQuoteCharacter": self._datasource_ascii_quote_character,
                "asciiThousandSeparator": self._datasource_ascii_thousand_separator,
                "dataSourceNameForClient": self._datasource_data_source_name_for_client,
                "dataSourceNameForServer": self._datasource_data_source_name_for_server
            }
            if self._datasource_ascii_delimiter_type == 'FixedWidth':
                del body_as_dict['DataSource']['asciiDelimiterChar']
        elif self._datasource_type == 'None':
            body_as_dict['DataSource'] = {
                "Type": "None"
            }
        elif self._datasource_type == 'ODBC':
            body_as_dict['DataSource'] = {
                "Type": self._datasource_type,
                "dataSourceNameForClient": self._datasource_data_source_name_for_client,
                "dataSourceNameForServer": self._datasource_data_source_name_for_server,
                "userName": self._datasource_user_name,
                "password": self._datasource_password,
                "query": self._datasource_query,
                "usesUnicode": self._datasource_uses_unicode
            }
        elif self._datasource_type == 'TM1CubeView':
            body_as_dict['DataSource'] = {
                "Type": self._datasource_type,
                "dataSourceNameForClient": self._datasource_data_source_name_for_server,
                "dataSourceNameForServer": self._datasource_data_source_name_for_server,
                "view": self._datasource_view
            }

        elif self._datasource_type == 'TM1DimensionSubset':
            body_as_dict['DataSource'] = {
                "Type": self._datasource_type,
                "dataSourceNameForClient": self._datasource_data_source_name_for_server,
                "dataSourceNameForServer": self._datasource_data_source_name_for_server,
                "subset": self._datasource_subset
            }
        return json.dumps(body_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-
import json
import warnings
from typing import List

from requests import Response

from TM1py.Exceptions.Exceptions import TM1pyException
from TM1py.Objects.Dimension import Dimension
from TM1py.Services.HierarchyService import HierarchyService
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.ProcessService import ProcessService
from TM1py.Services.RestService import RestService
from TM1py.Services.SubsetService import SubsetService
from TM1py.Utils.Utils import case_and_space_insensitive_equals, format_url, CaseAndSpaceInsensitiveSet


class DimensionService(ObjectService):
    """ Service to handle Object Updates for TM1 Dimensions
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)
        self.hierarchies = HierarchyService(rest)
        self.subsets = SubsetService(rest)

    def create(self, dimension: Dimension, **kwargs) -> Response:
        """ Create a dimension

        :param dimension: instance of TM1py.Dimension
        :return: response
        """
        # If Dimension exists. throw Exception
        if self.exists(dimension.name):
            raise RuntimeError("Dimension '{}' already exists".format(dimension.name))
        # If not all subsequent calls successful -> undo everything that has been done in this function
        try:
            # Create Dimension, Hierarchies, Elements, Edges.
            url = "/api/v1/Dimensions"
            response = self._rest.POST(url, dimension.body, **kwargs)
            # Create ElementAttributes
            for hierarchy in dimension:
                if not case_and_space_insensitive_equals(hierarchy.name, "Leaves"):
                    self.hierarchies.update_element_attributes(hierarchy, **kwargs)
        except TM1pyException as e:
            # undo everything if problem in step 1 or 2
            if self.exists(dimension.name, **kwargs):
                self.delete(dimension.name)
            raise e
        return response

    def get(self, dimension_name: str, **kwargs) -> Dimension:
        """ Get a Dimension

        :param dimension_name:
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')?$expand=Hierarchies($expand=*)", dimension_name)
        response = self._rest.GET(url, **kwargs)
        return Dimension.from_json(response.text)

    def update(self, dimension: Dimension, **kwargs):
        """ Update an existing dimension

        :param dimension: instance of TM1py.Dimension
        :return: None
        """
        # delete hierarchies that have been removed from the dimension object
        hierarchies_to_be_removed = CaseAndSpaceInsensitiveSet(
            *self.hierarchies.get_all_names(dimension.name, **kwargs))
        for hierarchy in dimension.hierarchy_names:
            hierarchies_to_be_removed.discard(hierarchy)

        # update all Hierarchies except for the implicitly maintained 'Leaves' Hierarchy
        for hierarchy in dimension:
            if not case_and_space_insensitive_equals(hierarchy.name, "Leaves"):
                if self.hierarchies.exists(hierarchy.dimension_name, hierarchy.name, **kwargs):
                    self.hierarchies.update(hierarchy, **kwargs)
                else:
                    self.hierarchies.create(hierarchy, **kwargs)

        # Edge case: elements in leaves hierarchy that do not exist in other hierarchies
        if "Leaves" in dimension:
            existing_leaves = CaseAndSpaceInsensitiveSet(
                self.hierarchies.elements.get_leaf_element_names(dimension.name, "Leaves"))

            leaves_to_create = list()
            for leaf in dimension.get_hierarchy("Leaves"):
                if leaf.name not in existing_leaves:
                    leaves_to_create.append(leaf)

            if leaves_to_create:
                self.hierarchies.elements.add_elements(
                    dimension_name=dimension.name,
                    hierarchy_name="Leaves",
                    elements=leaves_to_create)

        for hierarchy_name in hierarchies_to_be_removed:
            if not case_and_space_insensitive_equals(hierarchy_name, "Leaves"):
                self.hierarchies.delete(dimension_name=dimension.name, hierarchy_name=hierarchy_name, **kwargs)

    def update_or_create(self, dimension: Dimension, **kwargs):
        """ update if exists else create

        :param dimension:
        :return:
        """
        if self.exists(dimension_name=dimension.name, **kwargs):
            self.update(dimension=dimension, **kwargs)
        else:
            self.create(dimension=dimension, **kwargs)

    def delete(self, dimension_name: str, **kwargs) -> Response:
        """ Delete a dimension

        :param dimension_name: Name of the dimension
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')", dimension_name)
        return self._rest.DELETE(url, **kwargs)

    def exists(self, dimension_name: str, **kwargs) -> bool:
        """ Check if dimension exists
        
        :return: 
        """
        url = format_url("/api/v1/Dimensions('{}')", dimension_name)
        return self._exists(url, **kwargs)

    def get_all_names(self, skip_control_dims: bool = False, **kwargs) -> List[str]:
        """Ask TM1 Server for list of all dimension names

        :skip_control_dims: bool, True to skip control dims
        :Returns:
            List of Strings
        """
        url = format_url(
            "/api/v1/{}?$select=Name",
            'ModelDimensions()' if skip_control_dims else 'Dimensions'
        )

        response = self._rest.GET(url, **kwargs)

        dimension_names = list(entry['Name'] for entry in response.json()['value'])
        return dimension_names

    def get_number_of_dimensions(self, skip_control_dims: bool = False, **kwargs) -> int:
        """Ask TM1 Server for number of dimensions

        :skip_control_dims: bool, True to exclude control dims from count
        :return: Number of dimensions
        """

        if skip_control_dims:
            response = self._rest.GET("/api/v1/ModelDimensions()?$select=Name&$top=0&$count", **kwargs)
            return  response.json()['@odata.count']

        return int(self._rest.GET("/api/v1/Dimensions/$count", **kwargs).text)


    def execute_mdx(self, dimension_name: str, mdx: str, **kwargs) -> List:
        """ Execute MDX against Dimension. 
        Requires }ElementAttributes_ Cube of the dimension to exist !
 
        :param dimension_name: Name of the Dimension
        :param mdx: valid Dimension-MDX Statement 
        :return: List of Element names
        """

        warnings.warn("execute_mdx() will be deprecated; use ElementService execute_set_mdx.", DeprecationWarning,
                      stacklevel=2)

        mdx_skeleton = "SELECT " \
                       "{} ON ROWS, " \
                       "{{ [}}ElementAttributes_{}].DefaultMember }} ON COLUMNS  " \
                       "FROM [}}ElementAttributes_{}]"
        mdx_full = mdx_skeleton.format(mdx, dimension_name, dimension_name)
        request = '/api/v1/ExecuteMDX?$expand=Axes(' \
                  '$filter=Ordinal eq 1;' \
                  '$expand=Tuples($expand=Members($select=Ordinal;$expand=Element($select=Name))))'
        payload = {"MDX": mdx_full}
        response = self._rest.POST(request, json.dumps(payload, ensure_ascii=False), **kwargs)
        raw_dict = response.json()
        return [row_tuple['Members'][0]['Element']['Name'] for row_tuple in raw_dict['Axes'][0]['Tuples']]

    def create_element_attributes_through_ti(self, dimension: Dimension, **kwargs):
        """ 
        
        :param dimension. Instance of TM1py.Objects.Dimension class
        :return: 
        """
        process_service = ProcessService(self._rest)
        for h in dimension:
            statements = ["AttrInsert('{}', '', '{}', '{}');".format(dimension.name, ea.name, ea.attribute_type[0])
                          for ea
                          in h.element_attributes]
            process_service.execute_ti_code(lines_prolog=statements, **kwargs)



# -*- coding: utf-8 -*-
import json
from enum import Enum
from typing import List, Union, Iterable, Optional, Dict, Tuple

from mdxpy import MdxHierarchySet, Member, MdxLevelExpression
from requests import Response

from TM1py.Exceptions import TM1pyRestException, TM1pyException
from TM1py.Objects import ElementAttribute, Element
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import CaseAndSpaceInsensitiveDict, format_url, CaseAndSpaceInsensitiveSet, require_admin
from TM1py.Utils import build_element_unique_names, CaseAndSpaceInsensitiveTuplesDict


class MDXDrillMethod(Enum):
    TM1DRILLDOWNMEMBER = 1
    DESCENDANTS = 2


class ElementService(ObjectService):
    """ Service to handle Object Updates for TM1 Dimension (resp. Hierarchy) Elements

    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def get(self, dimension_name: str, hierarchy_name: str, element_name: str, **kwargs) -> Element:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')?$expand=*",
            dimension_name, hierarchy_name, element_name)
        response = self._rest.GET(url, **kwargs)
        return Element.from_dict(response.json())

    def create(self, dimension_name: str, hierarchy_name: str, element: Element, **kwargs) -> Response:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements",
            dimension_name,
            hierarchy_name)
        return self._rest.POST(url, element.body, **kwargs)

    def update(self, dimension_name: str, hierarchy_name: str, element: Element, **kwargs) -> Response:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
            dimension_name,
            hierarchy_name,
            element.name)
        return self._rest.PATCH(url, element.body, **kwargs)

    def exists(self, dimension_name: str, hierarchy_name: str, element_name: str, **kwargs) -> bool:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
            dimension_name,
            hierarchy_name,
            element_name)
        return self._exists(url, **kwargs)

    def delete(self, dimension_name: str, hierarchy_name: str, element_name: str, **kwargs) -> Response:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
            dimension_name,
            hierarchy_name,
            element_name)
        return self._rest.DELETE(url, **kwargs)

    def get_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[Element]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?select=Name,Type",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [Element.from_dict(element) for element in response.json()["value"]]

    def get_edges(self, dimension_name: str, hierarchy_name: str, **kwargs) -> Dict[Tuple[str, str], int]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Edges?select=ParentName,ComponentName,Weight",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)

        return {(edge["ParentName"], edge["ComponentName"]): edge["Weight"] for edge in response.json()["value"]}

    def get_leaf_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[Element]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$expand=*&$filter=Type ne 3",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [Element.from_dict(element) for element in response.json()["value"]]

    def get_leaf_element_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=Type ne 3",
                         dimension_name,
                         hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_consolidated_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[Element]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$expand=*&$filter=Type eq 3",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [Element.from_dict(element) for element in response.json()["value"]] 

    def get_consolidated_element_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=Type eq 3",
                         dimension_name,
                         hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_numeric_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[Element]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$expand=*&$filter=Type eq 1",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [Element.from_dict(element) for element in response.json()["value"]] 

    def get_numeric_element_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=Type eq 1",
                         dimension_name,
                         hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_string_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[Element]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$expand=*&$filter=Type eq 2",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [Element.from_dict(element) for element in response.json()["value"]] 

    def get_string_element_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=Type eq 2",
                         dimension_name,
                         hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_element_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        """ Get all element names

        :param dimension_name:
        :param hierarchy_name:
        :return: Generator of element-names
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_number_of_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements/$count",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_number_of_consolidated_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements/$count?$filter=Type eq 3",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_number_of_leaf_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements/$count?$filter=Type ne 3",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_number_of_numeric_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements/$count?$filter=Type eq 1",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_number_of_string_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements/$count?$filter=Type eq 2",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_all_leaf_element_identifiers(self, dimension_name: str, hierarchy_name: str,
                                         **kwargs) -> CaseAndSpaceInsensitiveSet:
        """ Get all element names and alias values for leaf elements in a hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """
        mdx_elements = f"{{ Tm1FilterByLevel ( {{ Tm1SubsetAll ([{dimension_name}].[{hierarchy_name}]) }} , 0 ) }}"
        return self.get_element_identifiers(dimension_name, hierarchy_name, mdx_elements, **kwargs)

    def get_elements_by_level(self, dimension_name: str, hierarchy_name: str, level: int, **kwargs) -> List[str]:
        """ Get all element names by level in a hierarchy

        :param dimension_name: Name of the dimension
        :param hierarchy_name: Name of the hierarchy
        :param level: Level to filter
        :return: List of element names
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=Level eq {}",
            dimension_name,
            hierarchy_name,
            str(level))
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_elements_filtered_by_wildcard(self, dimension_name: str, hierarchy_name: str,
                                          wildcard: str, level: int = None, **kwargs) -> List[str]:
        """ Get all element names filtered by wildcard (CaseAndSpaceInsensitive) and level in a hierarchy

        :param dimension_name: Name of the dimension
        :param hierarchy_name: Name of the hierarchy
        :param wildcard: wildcard to filter
        :param level: Level to filter
        :return: List of element names
        """
        filter_elements = format_url("contains(tolower(replace(Name,' ','')),tolower(replace('{}',' ', '')))", wildcard)
        if level is not None:
            filter_elements = filter_elements + f" and Level eq {level}"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name&$filter=" + filter_elements,
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [e["Name"] for e in response.json()['value']]

    def get_all_element_identifiers(self, dimension_name: str, hierarchy_name: str,
                                    **kwargs) -> CaseAndSpaceInsensitiveSet:
        """ Get all element names and alias values in a hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """

        mdx_elements = f"{{ Tm1SubsetAll ([{dimension_name}].[{hierarchy_name}]) }}"
        return self.get_element_identifiers(dimension_name, hierarchy_name, mdx_elements, **kwargs)

    def get_element_identifiers(self, dimension_name: str, hierarchy_name: str,
                                elements: Union[str, List[str]], **kwargs) -> CaseAndSpaceInsensitiveSet:
        """ Get all element names and alias values for a set of elements in a hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :param elements: MDX (Set) expression or iterable of elements
        :return:
        """

        alias_attributes = self.get_alias_element_attributes(dimension_name, hierarchy_name, **kwargs)

        if isinstance(elements, str):
            mdx_element_selection = elements
        else:
            mdx_element_selection = ",".join(build_element_unique_names(
                [dimension_name] * len(elements),
                elements,
                [hierarchy_name] * len(elements)))

        if not alias_attributes:
            result = self.execute_set_mdx(
                mdx=mdx_element_selection,
                member_properties=["Name"],
                parent_properties=None,
                element_properties=None)
            return CaseAndSpaceInsensitiveSet([record[0]["Name"] for record in result])

        mdx = """
             SELECT
             {{ {elem_mdx} }} ON ROWS, 
             {{ {attr_mdx} }} ON COLUMNS
             FROM [}}ElementAttributes_{dim}]
             """.format(
            elem_mdx=mdx_element_selection,
            attr_mdx=",".join(build_element_unique_names(
                ["}ElementAttributes_" + dimension_name] * len(alias_attributes), alias_attributes)),
            dim=dimension_name)
        return self._retrieve_mdx_rows_and_cell_values_as_string_set(mdx, **kwargs)

    def get_attribute_of_elements(self, dimension_name: str, hierarchy_name: str, attribute: str,
                                  elements: Union[str, List[str]] = None, exclude_empty_cells: bool = True,
                                  element_unique_names: bool = False) -> dict:
        """
         Get element name and attribute value for a set of elements in a hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :param attribute: Name of the Attribute
        :param elements:  MDX (Set) expression or iterable of elements
        :param exclude_empty_cells: Boolean
        :param element_unique_names: Boolean
        :return: Dict {'01':'Jan', '02':'Feb'}
        """
        if not elements:
            elements = self.get_element_names(dimension_name=dimension_name, hierarchy_name=hierarchy_name)

        if isinstance(elements, str):
            mdx_element_selection = elements
        else:
            mdx_element_selection = ",".join(build_element_unique_names(
                [dimension_name] * len(elements),
                elements,
                [hierarchy_name] * len(elements)))
        mdx = """
             SELECT
             {{ {elem_mdx} }} ON ROWS, 
             {{ {attr_mdx} }} ON COLUMNS
             FROM [}}ElementAttributes_{dim}]
             """.format(
            elem_mdx=mdx_element_selection,
            attr_mdx="[}ElementAttributes_" + dimension_name + "].[" + attribute + "]",
            dim=dimension_name)
        rows_and_values = self._retrieve_mdx_rows_and_values(mdx, element_unique_names=element_unique_names)
        return self._extract_dict_from_rows_and_values(rows_and_values, exclude_empty_cells=exclude_empty_cells)

    @staticmethod
    def _extract_dict_from_rows_and_values(
            rows_and_values: CaseAndSpaceInsensitiveTuplesDict,
            exclude_empty_cells: bool = True) -> dict:
        """ Helper function for get_element_by_attribute method

        :param rows_and_values:
        :param exclude_empty_cells: Boolean
        :return: Dictionary of Element:Attribute_Value
        """
        result_set = dict()
        for row_elements, cell_values in rows_and_values.items():
            for row_element in row_elements:
                for cell_value in cell_values:
                    if cell_value or not exclude_empty_cells:
                        result_set[row_element] = cell_value
        return result_set

    def get_level_names(self, dimension_name: str, hierarchy_name: str, descending: bool = True, **kwargs) -> List[str]:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Levels?$select=Name",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        if descending:
            return [level["Name"] for level in reversed(response.json()["value"])]
        else:
            return [level["Name"] for level in response.json()["value"]]

    def get_levels_count(self, dimension_name: str, hierarchy_name: str, **kwargs) -> int:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Levels/$count", dimension_name, hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def get_element_types(self, dimension_name: str, hierarchy_name: str,
                          skip_consolidations: bool = False, **kwargs) -> CaseAndSpaceInsensitiveDict:
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements?$select=Name,Type{}",
            dimension_name,
            hierarchy_name,
            "&$filter=Type ne 3" if skip_consolidations else "")
        response = self._rest.GET(url, **kwargs)

        result = CaseAndSpaceInsensitiveDict()
        for element in response.json()["value"]:
            result[element['Name']] = element["Type"]
        return result

    def get_element_types_from_all_hierarchies(
            self, dimension_name: str, skip_consolidations: bool = False, **kwargs) -> CaseAndSpaceInsensitiveDict:
        url = format_url(
            "/api/v1/Dimensions('{}')?$expand=Hierarchies($select=Elements;$expand=Elements($select=Name,Type{}",
            dimension_name,
            ";$filter=Type ne 3))" if skip_consolidations else "))")
        response = self._rest.GET(url, **kwargs)

        result = CaseAndSpaceInsensitiveDict()
        for hierarchy in response.json()["Hierarchies"]:
            for element in hierarchy["Elements"]:
                result[element['Name']] = element["Type"]
        return result

    def attribute_cube_exists(self, dimension_name: str, **kwargs) -> bool:
        url = format_url("/api/v1/Cubes('{}')", self.ELEMENT_ATTRIBUTES_PREFIX + dimension_name)
        return self._exists(url, **kwargs)

    def _retrieve_mdx_rows_and_cell_values_as_string_set(self, mdx: str, exclude_empty_cells=True, **kwargs):
        from TM1py import CellService
        return CellService(self._rest).execute_mdx_rows_and_values_string_set(mdx, exclude_empty_cells, **kwargs)

    def _retrieve_mdx_rows_and_values(self, mdx: str, **kwargs):
        from TM1py import CellService
        return CellService(self._rest).execute_mdx_rows_and_values(mdx, **kwargs)

    def get_alias_element_attributes(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        """

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """
        attributes = self.get_element_attributes(dimension_name, hierarchy_name, **kwargs)
        return [attr.name
                for attr
                in attributes if attr.attribute_type == 'Alias']

    def get_element_attributes(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[ElementAttribute]:
        """ Get element attributes from hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/ElementAttributes",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        element_attributes = [ElementAttribute.from_dict(ea) for ea in response.json()['value']]
        return element_attributes

    def get_element_attribute_names(self, dimension_name: str, hierarchy_name: str, **kwargs) -> List[str]:
        """ Get element attributes from hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/ElementAttributes?$select=Name",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return [ea["Name"] for ea in response.json()['value']]

    def get_elements_filtered_by_attribute(self, dimension_name: str, hierarchy_name: str, attribute_name: str,
                                           attribute_value: Union[str, float], **kwargs) -> List[str]:
        """ Get all elements from a hierarchy with given attribute value

        :param dimension_name:
        :param hierarchy_name:
        :param attribute_name:
        :param attribute_value:
        :return: List of element names
        """
        if not self.exists(f"}}ElementAttributes_{dimension_name}",
                           f"}}ElementAttributes_{dimension_name}",
                           attribute_name):
            raise RuntimeError(f"Attribute '{attribute_name}' does not exist in Dimension '{dimension_name}'")

        if isinstance(attribute_value, str):
            mdx = (
                f"{{FILTER({{TM1SUBSETALL([{dimension_name}].[{hierarchy_name}])}},"
                f"[{dimension_name}].[{hierarchy_name}].[{attribute_name}] = \"{attribute_value}\")}}")
        else:
            mdx = (
                f"{{FILTER({{TM1SUBSETALL([{dimension_name}].[{hierarchy_name}])}},"
                f"[{dimension_name}].[{hierarchy_name}].[{attribute_name}] = {attribute_value})}}")

        elems = self.execute_set_mdx(
            mdx=mdx,
            member_properties=["Name"],
            parent_properties=None,
            element_properties=None)
        return [elem[0]["Name"] for elem in elems]

    def create_element_attribute(self, dimension_name: str, hierarchy_name: str, element_attribute: ElementAttribute,
                                 **kwargs) -> Response:
        """ like AttrInsert

        :param dimension_name:
        :param hierarchy_name:
        :param element_attribute: instance of TM1py.ElementAttribute
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/ElementAttributes",
            dimension_name,
            hierarchy_name)
        return self._rest.POST(url, element_attribute.body, **kwargs)

    def delete_element_attribute(self, dimension_name: str, hierarchy_name: str, element_attribute: str,
                                 **kwargs) -> Response:
        """ like AttrDelete

        :param dimension_name:
        :param hierarchy_name:
        :param element_attribute: instance of TM1py.ElementAttribute
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('}}ElementAttributes_{}')/Hierarchies('}}ElementAttributes_{}')/Elements('{}')",
            dimension_name,
            hierarchy_name,
            element_attribute)
        try:
            return self._rest.DELETE(url, **kwargs)

        # Fail silently if attribute hierarchy or attribute doesn't exist
        except TM1pyRestException as ex:
            if not ex.status_code == 404:
                raise ex

    def get_leaves_under_consolidation(self, dimension_name: str, hierarchy_name: str, consolidation: str,
                                       max_depth: int = None, **kwargs) -> List[str]:
        """ Get all leaves under a consolidated element

        :param dimension_name: name of dimension
        :param hierarchy_name: name of hierarchy
        :param consolidation: name of consolidated Element
        :param max_depth: 99 if not passed
        :return:
        """
        return self.get_members_under_consolidation(dimension_name, hierarchy_name, consolidation, max_depth, True,
                                                    **kwargs)

    def get_members_under_consolidation(self, dimension_name: str, hierarchy_name: str, consolidation: str,
                                        max_depth: int = None, leaves_only: bool = False, **kwargs) -> List[str]:
        """ Get all members under a consolidated element

        :param dimension_name: name of dimension
        :param hierarchy_name: name of hierarchy
        :param consolidation: name of consolidated Element
        :param max_depth: 99 if not passed
        :param leaves_only: Only Leaf Elements or all Elements
        :return:
        """
        depth = max_depth - 1 if max_depth else 99
        # members to return
        members = []
        # build url
        bare_url = "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')?$select=Name,Type&$expand=Components("
        url = format_url(bare_url, dimension_name, hierarchy_name, consolidation)
        for _ in range(depth):
            url += "$select=Name,Type;$expand=Components("
        url = url[:-1] + ")" * depth

        response = self._rest.GET(url, **kwargs)
        consolidation_tree = response.json()

        # recursive function to parse consolidation_tree
        def get_members(element):
            if element["Type"] == "Numeric":
                members.append(element["Name"])
            elif element["Type"] == "Consolidated":
                if "Components" in element:
                    for component in element["Components"]:
                        if not leaves_only and component["Type"] == "Consolidated":
                            members.append(component["Name"])
                        get_members(component)

        get_members(consolidation_tree)
        return members

    def execute_set_mdx(
            self,
            mdx: str,
            top_records: Optional[int] = None,
            member_properties: Optional[Iterable[str]] = ('Name', 'Weight'),
            parent_properties: Optional[Iterable[str]] = ('Name', 'UniqueName'),
            element_properties: Optional[Iterable[str]] = ('Type', 'Level'),
            **kwargs) -> List:
        """
        :method to execute an MDX statement against a dimension
        :param mdx: valid dimension mdx statement
        :param top_records: number of records to return, default: all elements no limit
        :param member_properties: list of member properties (e.g., Name, UniqueName, Type, Weight, Attributes/Color)
        to return, will always return the Name property
        :param parent_properties: list of parent properties (e.g., Name, UniqueName, Type, Weight, Attributes/Color)
         to return, can be None or empty
        :param element_properties: list of element properties (e.g., Name, UniqueName, Type, Level, Index,
        Attributes/Color) to return, can be empty
        :return: dictionary of members, unique names, weights, types, and parents
        """

        top = f"$top={top_records};" if top_records else ""

        if not member_properties:
            member_properties = ['Name']

        # drop spaces in Attribute names
        else:
            member_properties = [
                member_property.replace(' ', '') if member_property.startswith('Attributes/') else member_property
                for member_property
                in member_properties]

        if element_properties:
            element_properties = [
                element_property.replace(' ', '') if element_property.startswith('Attributes/') else element_property
                for element_property
                in element_properties]

        if parent_properties:
            parent_properties = [
                parent_property.replace(' ', '') if parent_property.startswith('Attributes/') else parent_property
                for parent_property
                in parent_properties]

        member_properties = ",".join(member_properties)
        select_member_properties = f'$select={member_properties}'

        properties_to_expand = []
        if parent_properties:
            parent_properties = ",".join(parent_properties)
            select_parent_properties = f'Parent($select={parent_properties})'
            properties_to_expand.append(select_parent_properties)

        if element_properties:
            element_properties = ",".join(element_properties)
            select_element_properties = f'Element($select={element_properties})'
            properties_to_expand.append(select_element_properties)

        if properties_to_expand:
            expand_properties = f';$expand={",".join(properties_to_expand)}'
        else:
            expand_properties = ""

        url = f'/api/v1/ExecuteMDXSetExpression?$expand=Tuples({top}' \
              f'$expand=Members({select_member_properties}' \
              f'{expand_properties}))'

        payload = {"MDX": mdx}
        response = self._rest.POST(url, json.dumps(payload, ensure_ascii=False), **kwargs)
        raw_dict = response.json()
        return [tuples['Members'] for tuples in raw_dict['Tuples']]

    def remove_edge(self, dimension_name: str, hierarchy_name: str, parent: str, component: str, **kwargs) -> Response:
        """ Remove one edge from hierarchy. Fails if parent or child element doesn't exist.

        :param dimension_name:
        :param hierarchy_name:
        :param parent:
        :param component:
        :return:
        """

        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements('{}')/Edges(ParentName='{}',ComponentName='{}')",
            dimension_name,
            hierarchy_name,
            parent,
            parent,
            component)

        return self._rest.DELETE(url=url, **kwargs)

    def add_edges(self, dimension_name: str, hierarchy_name: str = None, edges: Dict[Tuple[str, str], int] = None,
                  **kwargs) -> Response:
        """ Add Edges to hierarchy. Fails if one edge already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param edges:
        :return:
        """
        if not hierarchy_name:
            hierarchy_name = dimension_name

        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Edges", dimension_name, hierarchy_name)
        body = [{"ParentName": parent, "ComponentName": component, "Weight": float(weight)}
                for (parent, component), weight
                in edges.items()]

        return self._rest.POST(url=url, data=json.dumps(body), **kwargs)

    def add_elements(self, dimension_name: str, hierarchy_name: str, elements: List[Element], **kwargs):
        """ Add elements to hierarchy. Fails if one element already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param elements:
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/Elements", dimension_name, hierarchy_name)
        body = [element.body_as_dict for element in elements]

        return self._rest.POST(url=url, data=json.dumps(body), **kwargs)

    def add_element_attributes(self, dimension_name: str, hierarchy_name: str,
                               element_attributes: List[ElementAttribute], **kwargs):
        """ Add element attributes to hierarchy. Fails if one element attribute already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param element_attributes:
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/ElementAttributes", dimension_name, hierarchy_name)
        body = [element_attribute.body_as_dict for element_attribute in element_attributes]

        return self._rest.POST(url=url, data=json.dumps(body), **kwargs)

    def get_parents(self, dimension_name: str, hierarchy_name: str, element_name: str, **kwargs) -> List[str]:
        url = format_url(
            f"/api/v1/Dimensions('{dimension_name}')/Hierarchies('{hierarchy_name}')/Elements('{element_name}')/Parents"
            f"?$select=Name",
        )
        response = self._rest.GET(url=url, **kwargs)

        return [record["Name"] for record in response.json()["value"]]

    def get_parents_of_all_elements(self, dimension_name: str, hierarchy_name: str, **kwargs) -> Dict[str, List[str]]:
        url = format_url(
            f"/api/v1/Dimensions('{dimension_name}')/Hierarchies('{hierarchy_name}')/Elements?$select=Name"
            f"&$expand=Parents($select=Name)",
        )
        response = self._rest.GET(url=url, **kwargs)

        return {child["Name"]: [parent["Name"] for parent in child["Parents"]] for child in response.json()["value"]}

    def get_element_principal_name(self, dimension_name: str, hierarchy_name: str, element_name: str, **kwargs) -> str:
        element = self.get(dimension_name, hierarchy_name, element_name, **kwargs)
        return element.name

    def _get_mdx_set_cardinality(self, mdx: str) -> int:
        url = format_url("/api/v1/ExecuteMDXSetExpression?$select=Cardinality")
        payload = {"MDX": mdx}
        response = self._rest.POST(url, json.dumps(payload, ensure_ascii=False))
        return response.json()['Cardinality']

    @staticmethod
    def _build_drill_intersection_mdx(dimension_name: str, hierarchy_name: str, first_element_name: str,
                                      second_element_name: str, mdx_method: str, recursive: bool) -> str:

        first_member = Member.of(dimension_name, hierarchy_name, first_element_name)
        second_member = Member.of(dimension_name, hierarchy_name, second_element_name)

        if MDXDrillMethod.TM1DRILLDOWNMEMBER.name == mdx_method.upper():
            query = MdxHierarchySet.members([first_member]).tm1_drill_down_member(recursive=recursive)

        elif MDXDrillMethod.DESCENDANTS.name == mdx_method.upper():
            query = MdxHierarchySet.descendants(
                member=first_member,
                level_or_depth=MdxLevelExpression.member_level(second_member),
                desc_flag='SELF')

        else:
            raise TM1pyException("Invalid MDX Drill Method Specified, Options: 'TM1DrillDownMember' or 'Descendants'")

        mdx = query.intersect(MdxHierarchySet.members([second_member]))
        return mdx.to_mdx()

    def element_is_parent(self, dimension_name: str, hierarchy_name: str, parent_name: str,
                          element_name: str) -> bool:
        """ Element is Parent
        :Note, unlike the related function in TM1 (ELISPAR or ElementIsParent), this function will return False
        :if an invalid element is passed;
        :but will raise an exception if an invalid dimension, or hierarchy is passed
        """
        mdx = self._build_drill_intersection_mdx(dimension_name=dimension_name,
                                                 hierarchy_name=hierarchy_name,
                                                 first_element_name=parent_name,
                                                 second_element_name=element_name,
                                                 mdx_method='TM1DrillDownMember',
                                                 recursive=False)

        cardinality = self._get_mdx_set_cardinality(mdx)
        return bool(cardinality)

    def element_is_ancestor(self, dimension_name: str, hierarchy_name: str, ancestor_name: str,
                            element_name: str, method: str = None) -> bool:
        """ Element is Ancestor

        :Note, unlike the related function in TM1 (`ELISANC` or `ElementIsAncestor`), this function will return False
        if an invalid element is passed; but will raise an exception if an invalid dimension, or hierarchy is passed

        For `method` you can pass 3 three values
        value `TI` performs best, but requires admin permissions
        Value 'TM1DrillDownMember' performs well when element is a leaf.
        Value 'Descendants' performs well when `ancestor_name` and `element_name` are Consolidations.

        If no value is passed, function defaults to 'TI' for user with admin permissions
        and 'TM1DrillDownMember' for users without admin permissions
        """
        if not method:
            method = 'TI' if self.is_admin else 'TM1DrillDownMember'

        if method.upper() == "TI":
            if self._element_is_ancestor_ti(dimension_name, hierarchy_name, element_name, ancestor_name):
                return True

            if self.hierarchy_exists(dimension_name, hierarchy_name):
                return False

            raise TM1pyException(f"Hierarchy: '{hierarchy_name}' does not exist in dimension: '{dimension_name}'")

        # make sure DESCENDANTS behaves like default TM1DrillDownMember
        if method.upper() == MDXDrillMethod.DESCENDANTS.name:
            if not self.exists(dimension_name, hierarchy_name, element_name):

                # case dimension or hierarchy doesn't exist
                if not self.hierarchy_exists(dimension_name, hierarchy_name):
                    raise TM1pyException(f"Hierarchy '{hierarchy_name}' does not exist in dimension '{dimension_name}'")

                # case element doesn't exist
                return False

        mdx = self._build_drill_intersection_mdx(
            dimension_name=dimension_name,
            hierarchy_name=hierarchy_name,
            first_element_name=ancestor_name,
            second_element_name=element_name,
            mdx_method=method,
            recursive=True)

        cardinality = self._get_mdx_set_cardinality(mdx)
        return bool(cardinality)

    def hierarchy_exists(self, dimension_name, hierarchy_name):
        hierarchy_service = self._get_hierarchy_service()
        return hierarchy_service.exists(dimension_name, hierarchy_name)

    @require_admin
    def _element_is_ancestor_ti(self, dimension_name: str, hierarchy_name: str, element_name: str,
                                ancestor_name: str) -> bool:
        process_service = self.get_process_service()
        code = f"ElementIsAncestor('{dimension_name}', '{hierarchy_name}', '{ancestor_name}', '{element_name}')=1"
        return process_service.evaluate_boolean_ti_expression(code)


    def get_process_service(self):
        from TM1py import ProcessService
        return ProcessService(self._rest)

    def _get_hierarchy_service(self):
        from TM1py import HierarchyService
        return HierarchyService(self._rest)



# -*- coding: utf-8 -*-
import json
from enum import Enum
from typing import Dict, Union

from TM1py.Objects.TM1Object import TM1Object


class HitMode(Enum):
    BREAK_ALWAYS = 'BreakAlways'
    BREAK_EQUAL = 'BreakEqual'
    BREAK_GREATER_OR_EQUAL = 'BreakGreaterOrEqual'

    def __str__(self):
        return self.value

    @classmethod
    def _missing_(cls, value: str):
        for member in cls:
            if member.value.lower() == value.replace(" ", "").lower():
                return member

        # default
        raise ValueError(f"Invalid HitMode: '{value}'")


class BreakPointType(Enum):
    # A breakpoint that pauses execution when the named variable is written to
    PROCESS_DEBUG_CONTEXT_DATA_BREAK_POINT = "ProcessDebugContextDataBreakpoint"
    # A breakpoint that pauses execution at a specific script line
    PROCESS_DEBUG_CONTEXT_LINE_BREAK_POINT = "ProcessDebugContextLineBreakpoint"
    # A breakpoint that pauses execution when an object lock is acquired.
    PROCESS_DEBUG_CONTEXT_LOCK_BREAK_POINT = "ProcessDebugContextLockBreakpoint"

    def __str__(self):
        return self.value

    @classmethod
    def _missing_(cls, value: str):
        for member in cls:
            if member.value.lower() == value.replace(" ", "").lower():
                return member

        # default
        raise ValueError(f"Invalid BreakPointType: '{value}'")


class ProcessDebugBreakpoint(TM1Object):
    """ Abstraction of a TM1 Process Debug Breakpoint.

    """

    def __init__(
            self,
            breakpoint_id: int,
            breakpoint_type: Union[BreakPointType, str] = BreakPointType.PROCESS_DEBUG_CONTEXT_LINE_BREAK_POINT,
            enabled: bool = True,
            hit_mode: Union[HitMode, str] = HitMode.BREAK_ALWAYS,
            hit_count: int = 0,
            expression: str = '',
            variable_name: str = '',
            process_name: str = '',
            procedure: str = '',
            line_number: int = 0,
            object_name: str = '',
            object_type: str = '',
            lock_mode: str = ''
    ):
        self._type = BreakPointType(breakpoint_type)
        self._id = breakpoint_id
        self._enabled = enabled
        self._hit_mode = HitMode(hit_mode)
        self._hit_count = hit_count
        self._expression = expression
        self._variable_name = variable_name
        self._process_name = process_name
        self._procedure = procedure
        self._line_number = line_number
        self._object_name = object_name
        self._object_type = object_type
        self._lock_mode = lock_mode

    @classmethod
    def from_dict(cls, breakpoint_as_dict: Dict) -> 'ProcessDebugBreakpoint':
        """

        :param breakpoint_as_dict
        :return: an instance of this class
        """
        breakpoint_type = breakpoint_as_dict['@odata.type'][16:]
        return cls(
            breakpoint_type=breakpoint_type,
            breakpoint_id=breakpoint_as_dict['ID'],
            enabled=breakpoint_as_dict['Enabled'],
            hit_mode=breakpoint_as_dict['HitMode'],
            hit_count=breakpoint_as_dict['HitCount'],
            expression=breakpoint_as_dict['Expression'],
            variable_name=breakpoint_as_dict[
                'VariableName'] if breakpoint_type == "ProcessDebugContextDataBreakpoint" else "",
            process_name=breakpoint_as_dict[
                'ProcessName'] if breakpoint_type == "ProcessDebugContextLineBreakpoint" else "",
            procedure=breakpoint_as_dict['Procedure'] if breakpoint_type == "ProcessDebugContextLineBreakpoint" else "",
            line_number=breakpoint_as_dict[
                'LineNumber'] if breakpoint_type == "ProcessDebugContextLineBreakpoint" else "",
            object_name=breakpoint_as_dict[
                'ObjectName'] if breakpoint_type == "ProcessDebugContextLockBreakpoint" else "",
            object_type=breakpoint_as_dict[
                'ObjectType'] if breakpoint_type == "ProcessDebugContextLockBreakpoint" else "",
            lock_mode=breakpoint_as_dict['LockMode'] if breakpoint_type == "ProcessDebugContextLockBreakpoint" else ""
        )

    @property
    def breakpoint_type(self) -> str:
        return str(self._type)

    @property
    def breakpoint_id(self) -> int:
        return self._id

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def hit_mode(self) -> str:
        return str(self._hit_mode)

    @property
    def hit_count(self) -> int:
        return self._hit_count

    @property
    def expression(self) -> str:
        return self._expression

    @property
    def variable_name(self) -> str:
        return self._variable_name

    @property
    def process_name(self) -> str:
        return self._process_name

    @property
    def procedure(self) -> str:
        return self._procedure

    @property
    def line_number(self) -> int:
        return self._line_number

    @property
    def object_name(self) -> str:
        return self._object_name

    @property
    def object_type(self) -> str:
        return self._object_type

    @property
    def lock_mode(self) -> str:
        return self._lock_mode

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body())

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value

    @hit_mode.setter
    def hit_mode(self, value: Union[HitMode, str]):
        self._hit_mode = HitMode(value)

    @expression.setter
    def expression(self, value: str):
        self._expression = value

    @variable_name.setter
    def variable_name(self, value: str):
        self._variable_name = value

    @process_name.setter
    def process_name(self, value: str):
        self._process_name = value

    @procedure.setter
    def procedure(self, value: str):
        self._procedure = value

    @line_number.setter
    def line_number(self, value: str):
        self._line_number = value

    @object_name.setter
    def object_name(self, value: str):
        self._object_name = value

    @object_type.setter
    def object_type(self, value: str):
        self._object_type = value

    @lock_mode.setter
    def lock_mode(self, value: str):
        self._lock_mode = value

    def _construct_body(self) -> str:
        body_as_dict = {
            '@odata.type': "#ibm.tm1.api.v1." + str(self._type),
            'ID': self._id,
            'Enabled': self._enabled,
            'HitMode': str(self._hit_mode),
            'Expression': self._expression
        }

        if self._type == BreakPointType.PROCESS_DEBUG_CONTEXT_DATA_BREAK_POINT:
            body_as_dict['VariableName'] = self._variable_name

        elif self._type == BreakPointType.PROCESS_DEBUG_CONTEXT_LINE_BREAK_POINT:
            body_as_dict['ProcessName'] = self._process_name
            body_as_dict['Procedure'] = self._procedure
            body_as_dict['LineNumber'] = self._line_number

        elif self._type == BreakPointType.PROCESS_DEBUG_CONTEXT_LOCK_BREAK_POINT:
            body_as_dict['ObjectName'] = self._object_name
            body_as_dict['ObjectType'] = self._object_type
            body_as_dict['LockMode'] = self._lock_mode

        return body_as_dict



# -*- coding: utf-8 -*-
from typing import List


class GitRemote:
    """ Abstraction of GitRemote
    """

    def __init__(self, connected: bool, branches: List[str], tags: List[str]):
        """ Initialize GitRemote object
        :param connected: is Git connected to remote
        :param branches: list of remote branches
        :param tags: list of remote tags
        """
        self._connected = connected
        self._branches = branches
        self._tags = tags

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def branches(self) -> List[str]:
        return self._branches

    @property
    def tags(self) -> List[str]:
        return self._tags



from TM1py.Objects.Annotation import Annotation
from TM1py.Objects.Application import Application
from TM1py.Objects.Axis import ViewAxisSelection, ViewTitleSelection
from TM1py.Objects.Chore import Chore
from TM1py.Objects.ChoreFrequency import ChoreFrequency
from TM1py.Objects.ChoreStartTime import ChoreStartTime
from TM1py.Objects.ChoreTask import ChoreTask
from TM1py.Objects.Cube import Cube
from TM1py.Objects.Dimension import Dimension
from TM1py.Objects.Element import Element
from TM1py.Objects.ElementAttribute import ElementAttribute
from TM1py.Objects.GitProject import TM1Project, TM1ProjectTask
from TM1py.Objects.Hierarchy import Hierarchy
from TM1py.Objects.MDXView import MDXView
from TM1py.Objects.NativeView import NativeView
from TM1py.Objects.Process import Process
from TM1py.Objects.ProcessDebugBreakpoint import ProcessDebugBreakpoint, BreakPointType, HitMode
from TM1py.Objects.Rules import Rules
from TM1py.Objects.Server import Server
from TM1py.Objects.Subset import Subset, AnonymousSubset
from TM1py.Objects.User import User
from TM1py.Objects.View import View
from TM1py.Objects.Sandbox import Sandbox



# -*- coding: utf-8 -*-

import collections
import json
from typing import Iterable, List, Dict, Optional, Union

from TM1py.Objects.Rules import Rules
from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url


class Cube(TM1Object):
    """ Abstraction of a TM1 Cube
        
    """

    def __init__(self, name: str, dimensions: Iterable[str], rules: Optional[Union[str, Rules]] = None):
        """
        
        :param name: name of the Cube
        :param dimensions: list of (existing) dimension names
        :param rules: instance of TM1py.Objects.Rules
        """
        self._name = name
        self.dimensions = list(dimensions)
        self.rules = rules

    @property
    def name(self) -> str:
        return self._name

    @property
    def dimensions(self) -> List[str]:
        return self._dimensions

    @dimensions.setter
    def dimensions(self, value: List[str]):
        self._dimensions = value

    @property
    def has_rules(self) -> bool:
        if self._rules:
            return True
        return False

    @property
    def rules(self) -> Rules:
        return self._rules

    @rules.setter
    def rules(self, value: Union[str, Rules]):
        if value is None:
            self._rules = None
        elif isinstance(value, str):
            self._rules = Rules(rules=value)
        elif isinstance(value, Rules):
            self._rules = value
        else:
            raise ValueError('value must None or of type str or Rules')

    @property
    def skipcheck(self) -> bool:
        if self.has_rules:
            return self.rules.skipcheck
        return False

    @property
    def undefvals(self) -> bool:
        if self.has_rules:
            return self.rules.undefvals
        return False

    @property
    def feedstrings(self) -> bool:
        if self.has_rules:
            return self.rules.feedstrings
        return False

    @classmethod
    def from_json(cls, cube_as_json: str) -> 'Cube':
        """ Alternative constructor

        :param cube_as_json: user as JSON string
        :return: cube, an instance of this class
        """
        cube_as_dict = json.loads(cube_as_json)
        return cls.from_dict(cube_as_dict)

    @classmethod
    def from_dict(cls, cube_as_dict: Dict) -> 'Cube':
        """ Alternative constructor

        :param cube_as_dict: user as dict
        :return: user, an instance of this class
        """
        return cls(
            name=cube_as_dict['Name'],
            dimensions=[dimension['Name'] for dimension in cube_as_dict['Dimensions']],
            rules=Rules(cube_as_dict['Rules']) if cube_as_dict['Rules'] else None)

    @property
    def body(self) -> str:
        return self._construct_body()

    def _construct_body(self) -> str:
        """
        construct body (json) from the class attributes
        :return: String, TM1 JSON representation of a cube
        """
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self.name
        body_as_dict['Dimensions@odata.bind'] = [format_url("Dimensions('{}')", dimension)
                                                 for dimension
                                                 in self.dimensions]
        if self.has_rules:
            body_as_dict['Rules'] = str(self.rules)
        return json.dumps(body_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-
import functools
import re
import time
import warnings
from base64 import b64encode, b64decode
from http.client import HTTPResponse
from io import BytesIO
from typing import Union, Dict, Tuple, Optional

import requests
import urllib3
import socket
from requests import Timeout, Response, ConnectionError
from requests.adapters import HTTPAdapter

# SSO not supported for Linux
from TM1py.Exceptions.Exceptions import TM1pyTimeout
from TM1py.Utils import case_and_space_insensitive_equals, CaseAndSpaceInsensitiveSet, HTTPAdapterWithSocketOptions

try:
    from requests_negotiate_sspi import HttpNegotiateAuth
except ImportError:
    warnings.warn("requests_negotiate_sspi failed to import. SSO will not work", ImportWarning)

from TM1py.Exceptions import TM1pyRestException

import http.client as http_client


def httpmethod(func):
    """ Higher Order Function to wrap the GET, POST, PATCH, PUT, DELETE methods
        Takes care of:
        - encoding of url and payload
        - verifying response. Throws TM1pyException if StatusCode of Response is not OK
    """

    @functools.wraps(func)
    def wrapper(self, url: str, data: str = '', encoding='utf-8', async_requests_mode: Optional[bool] = None, tcp_keepalive: Optional[bool] = None, **kwargs):
        # url encoding
        url, data = self._url_and_body(
            url=url,
            data=data,
            encoding=encoding)
        # execute request
        try:
            # determine async_requests_mode
            if async_requests_mode is None:
                async_requests_mode = self._async_requests_mode

            if tcp_keepalive is None:
                tcp_keepalive = self._tcp_keepalive

            if not async_requests_mode:
                response = func(self, url, data, **kwargs)

            else:
                # reset tcp_keepalive to False explicitly to turn it off when async_requests_mode is enabled
                tcp_keepalive = False
                additional_header = {'Prefer': 'respond-async'}
                http_headers = kwargs.get('headers', dict())
                http_headers.update(additional_header)
                kwargs['headers'] = http_headers
                response = func(self, url, data, **kwargs)
                self.verify_response(response=response)

                if 'Location' not in response.headers or "'" not in response.headers['Location']:
                    raise ValueError(f"Failed to retrieve async_id from request {func.__name__} '{url}'")
                async_id = response.headers.get('Location').split("'")[1]

                for wait in RestService.wait_time_generator(kwargs.get('timeout', self._timeout)):
                    response = self.retrieve_async_response(async_id)
                    if response.status_code in [200, 201]:
                        break
                    time.sleep(wait)

                # all wait times consumed and still no 200
                if response.status_code not in [200, 201]:
                    if kwargs.get("cancel_at_timeout", False):
                        self.cancel_async_operation(async_id)
                    raise TM1pyTimeout(method=func.__name__, url=url, timeout=kwargs['timeout'])

                response = self.build_response_from_raw_bytes(response.content)

            # verify
            self.verify_response(response=response)

            # response encoding
            response.encoding = encoding
            return response

        except Timeout:
            if kwargs.get("cancel_at_timeout", False):
                self.cancel_running_operation()
            raise TM1pyTimeout(method=func.__name__, url=url, timeout=kwargs['timeout'])

        except ConnectionError as e:
            # cater for issue in requests library: https://github.com/psf/requests/issues/5430
            if re.search('Read timed out', str(e), re.IGNORECASE):
                if kwargs.get("cancel_at_timeout", False):
                    self.cancel_running_operation()
                raise TM1pyTimeout(method=func.__name__, url=url, timeout=kwargs['timeout'])

    return wrapper


class RestService:
    """ Low level communication with TM1 instance through HTTP.
        Allows to execute HTTP Methods
            - GET
            - POST
            - PATCH
            - DELETE
        Takes Care of
            - Encodings
            - TM1 User-Login
            - HTTP Headers
            - HTTP Session Management
            - Response Handling
        Based on requests module
    """

    HEADERS = {'Connection': 'keep-alive',
               'User-Agent': 'TM1py',
               'Content-Type': 'application/json; odata.streaming=true; charset=utf-8',
               'Accept': 'application/json;odata.metadata=none,text/plain',
               'TM1-SessionContext': 'TM1py'}

    # You can reset the following TCP socket options based on your own use cases when tcp_keepalive is eanbled
    # TCP_KEEPIDLE: Time in seconds until the first keepalive is sent
    # TCP_KEEPINTVL: How often should the keepalive packet be sent
    # TCP_KEEPCNT: The max number of keepalive packets to send
    TCP_SOCKET_OPTIONS = {'TCP_KEEPIDLE': 30,
                          'TCP_KEEPINTVL': 15,
                          'TCP_KEEPCNT': 60}

    def __init__(self, **kwargs):
        """ Create an instance of RESTService
        :param address: String - address of the TM1 instance
        :param port: Int - HTTPPortNumber as specified in the tm1s.cfg
        :param base_url - base url e.g. https://localhost:12354/api/v1
        :param user: String - name of the user
        :param password String - password of the user
        :param decode_b64 - whether password argument is b64 encoded
        :param namespace String - optional CAM namespace
        :param ssl: boolean -  as specified in the tm1s.cfg
        :param cam_passport: String - the cam passport
        :param session_id: String - TM1SessionId e.g. q7O6e1w49AixeuLVxJ1GZg
        :param session_context: String - Name of the Application. Controls "Context" column in Arc / TM1top.
        If None, use default: TM1py
        :param verify: path to .cer file or 'True' / True / 'False' / False (if no ssl verification is required)
        :param logging: boolean - switch on/off verbose http logging into sys.stdout
        :param timeout: Float - Number of seconds that the client will wait to receive the first byte.
        :param async_requests_mode: changes internal REST execution mode to avoid 60s timeout on IBM cloud
        :param tcp_keepalive: maintain the TCP connection all the time, users should choose either async_requests_mode or tcp_keepalive to run a long-run request
        If both are True, use async_requests_mode by default
        :param connection_pool_size - In a multithreaded environment, you should set this value to a
        higher number, such as the number of threads
        :param integrated_login: True for IntegratedSecurityMode3
        :param integrated_login_domain: NT Domain name.
                Default: '.' for local account.
        :param integrated_login_service: Kerberos Service type for remote Service Principal Name.
                Default: 'HTTP'
        :param integrated_login_host: Host name for Service Principal Name.
                Default: Extracted from request URI
        :param integrated_login_delegate: Indicates that the user's credentials are to be delegated to the server.
                Default: False
        :param impersonate: Name of user to impersonate
        """
        self._kwargs = kwargs
        self._ssl = self.translate_to_boolean(kwargs.get('ssl', True))
        self._address = kwargs.get('address', None)
        self._port = kwargs.get('port', None)
        self._verify = False
        self._timeout = None if kwargs.get('timeout', None) is None else float(kwargs.get('timeout'))
        self._async_requests_mode = self.translate_to_boolean(kwargs.get('async_requests_mode', False))
        self._tcp_keepalive = self.translate_to_boolean(kwargs.get('tcp_keepalive', False))
        self._connection_pool_size = kwargs.get('connection_pool_size', None)
        # populated on the fly
        if kwargs.get('user'):
            self._is_admin = True if case_and_space_insensitive_equals(kwargs.get('user'), 'ADMIN') else None
        else:
            self._is_admin = None

        if 'verify' in kwargs:
            if isinstance(kwargs['verify'], str):
                if kwargs['verify'].upper() == 'FALSE':
                    self._verify = False
                elif kwargs['verify'].upper() == 'TRUE':
                    self._verify = True
                # path to .cer file
                else:
                    self._verify = kwargs.get('verify')
            elif isinstance(kwargs['verify'], bool):
                self._verify = kwargs['verify']
            else:
                raise ValueError("verify argument must be of type str or bool")

        if 'base_url' in kwargs:
            self._base_url = kwargs['base_url']
            self._ssl = self._determine_ssl_based_on_base_url()
        else:
            self._base_url = "http{}://{}:{}".format(
                's' if self._ssl else '',
                'localhost' if len(self._address) == 0 else self._address,
                self._port)

        self._version = None
        self._headers = self.HEADERS.copy()
        if "session_context" in kwargs:
            self._headers["TM1-SessionContext"] = kwargs["session_context"]

        self.disable_http_warnings()
        # re-use or create tm1 http session
        self._s = requests.session()
        if "session_id" in kwargs:
            self._s.cookies.set("TM1SessionId", kwargs["session_id"])
        else:
            self._start_session(
                user=kwargs.get("user", None),
                password=kwargs.get("password", None),
                namespace=kwargs.get("namespace", None),
                gateway=kwargs.get("gateway", None),
                cam_passport=kwargs.get("cam_passport", None),
                decode_b64=self.translate_to_boolean(kwargs.get("decode_b64", False)),
                integrated_login=self.translate_to_boolean(kwargs.get("integrated_login", False)),
                integrated_login_domain=kwargs.get("integrated_login_domain"),
                integrated_login_service=kwargs.get("integrated_login_service"),
                integrated_login_host=kwargs.get("integrated_login_host"),
                integrated_login_delegate=kwargs.get("integrated_login_delegate"),
                impersonate=kwargs.get("impersonate", None))

        if not self._version:
            self.set_version()

        self._sandboxing_disabled = None

        if self._tcp_keepalive or self._connection_pool_size is not None:
            self._manage_http_adapter()

        # Logging
        if 'logging' in kwargs:
            if self.translate_to_boolean(value=kwargs['logging']):
                http_client.HTTPConnection.debuglevel = 1

    def _manage_http_adapter(self):
        if self._tcp_keepalive:
            # SO_KEEPALIVE: set 1 to enable TCP keepalive
            socket_options = urllib3.connection.HTTPConnection.default_socket_options + [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                                                                                         (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, self.TCP_SOCKET_OPTIONS['TCP_KEEPIDLE']),
                                                                                         (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, self.TCP_SOCKET_OPTIONS['TCP_KEEPINTVL']),
                                                                                         (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, self.TCP_SOCKET_OPTIONS['TCP_KEEPCNT'])]

            if self._connection_pool_size is not None:
                adapter = HTTPAdapterWithSocketOptions(pool_connections=int(self._connection_pool_size),
                                                       pool_maxsize=int(self._connection_pool_size),
                                                       socket_options=socket_options)
            else:
                adapter = HTTPAdapterWithSocketOptions(socket_options=socket_options)

        else:
            adapter = HTTPAdapterWithSocketOptions(pool_connections=int(self._connection_pool_size),
                                                   pool_maxsize=int(self._connection_pool_size))

        self._s.mount(self._base_url, adapter)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.logout()

    @httpmethod
    def GET(self, url: str, data: Union[str, bytes] = '', headers: Dict = None, timeout: float = None, **kwargs):
        """ Perform a GET request against TM1 instance
        :param url:
        :param data: the payload
        :param headers: custom headers
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :return: response object
        """
        return self._s.get(
            url=url,
            headers={**self._headers, **headers} if headers else self._headers,
            data=data,
            verify=self._verify,
            timeout=timeout if timeout else self._timeout)

    @httpmethod
    def POST(self, url: str, data: Union[str, bytes], headers: Dict = None, timeout: float = None, **kwargs):
        """ POST request against the TM1 instance
        :param url:
        :param data: the payload
        :param headers: custom headers
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :return: response object
        """
        return self._s.post(
            url=url,
            headers={**self._headers, **headers} if headers else self._headers,
            data=data,
            verify=self._verify,
            timeout=timeout if timeout else self._timeout)

    @httpmethod
    def PATCH(self, url: str, data: Union[str, bytes], headers: Dict = None, timeout: float = None, **kwargs):
        """ PATCH request against the TM1 instance
        :param url: String, for instance : /api/v1/Dimensions('plan_business_unit')
        :param data: the payload
        :param headers: custom headers
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :return: response object
        """
        return self._s.patch(
            url=url,
            headers={**self._headers, **headers} if headers else self._headers,
            data=data,
            verify=self._verify,
            timeout=timeout if timeout else self._timeout)

    @httpmethod
    def PUT(self, url: str, data: Union[str, bytes], headers: Dict = None, timeout: float = None, **kwargs):
        """ PUT request against the TM1 instance
        :param url: String, for instance : /api/v1/Dimensions('plan_business_unit')
        :param data: the payload
        :param headers: custom headers
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :return: response object
        """
        return self._s.put(
            url=url,
            headers={**self._headers, **headers} if headers else self._headers,
            data=data,
            verify=self._verify,
            timeout=timeout if timeout else self._timeout)

    @httpmethod
    def DELETE(self, url: str, data: Union[str, bytes], headers: Dict = None, timeout: float = None, **kwargs):
        """ Delete request against TM1 instance
        :param url:  String, for instance : /api/v1/Dimensions('plan_business_unit')
        :param data: the payload
        :param headers: custom headers
        :param timeout: Number of seconds that the client will wait to receive the first byte.
        :return: response object
        """
        return self._s.delete(
            url=url,
            headers={**self._headers, **headers} if headers else self._headers,
            data=data,
            verify=self._verify,
            timeout=timeout if timeout else self._timeout)

    def logout(self, timeout: float = None, **kwargs):
        """ End TM1 Session and HTTP session
        """
        # Easier to ask for forgiveness than permission
        try:
            # ProductVersion >= TM1 10.2.2 FP 6
            self.POST('/api/v1/ActiveSession/tm1.Close', '', headers={"Connection": "close"}, timeout=timeout,
                      async_requests_mode=False, **kwargs)
        except TM1pyRestException:
            # ProductVersion < TM1 10.2.2 FP 6
            self.POST('/api/logout', '', headers={"Connection": "close"}, timeout=timeout, **kwargs)
        finally:
            self._s.close()

    def _start_session(self, user: str, password: str, decode_b64: bool = False, namespace: str = None,
                       gateway: str = None, cam_passport: str = None, integrated_login: bool = None,
                       integrated_login_domain: str = None, integrated_login_service: str = None,
                       integrated_login_host: str = None, integrated_login_delegate: bool = None,
                       impersonate: str = None):
        """ perform a simple GET request (Ask for the TM1 Version) to start a session
        """
        # Authorization with integrated_login
        if integrated_login:
            self._s.auth = HttpNegotiateAuth(
                domain=integrated_login_domain,
                service=integrated_login_service,
                host=integrated_login_host,
                delegate=integrated_login_delegate)

        # Authorization [Basic, CAM] through Headers
        else:
            token = self._build_authorization_token(
                user,
                self.b64_decode_password(password) if decode_b64 else password,
                namespace,
                gateway,
                cam_passport,
                self._verify)
            self.add_http_header('Authorization', token)

        url = '/api/v1/Configuration/ProductVersion/$value'
        try:
            additional_headers = dict()
            if impersonate:
                additional_headers["TM1-Impersonate"] = impersonate

            response = self.GET(url=url, headers=additional_headers)
            if response is None:
                raise ValueError(f"No response returned from URL: '{self._base_url + url}'. "
                                 f"Please double check your address and port number in the URL.")

            self._version = response.text
        finally:
            # After we have session cookie, drop the Authorization Header
            self.remove_http_header('Authorization')

    def _url_and_body(self, url: str, data: str, encoding: str = 'utf-8') -> Tuple[str, bytes]:
        """ create proper url and payload
        """
        url = self._base_url + url
        url = url.replace(' ', '%20')
        if isinstance(data, str):
            data = data.encode(encoding)
        return url, data

    def is_connected(self) -> bool:
        """ Check if Connection to TM1 Server is established.
        :Returns:
            Boolean
        """
        try:
            self.GET('/api/v1/Configuration/ServerName/$value')
            return True
        except:
            return False

    def set_version(self):
        url = '/api/v1/Configuration/ProductVersion/$value'
        response = self.GET(url=url)
        self._version = response.text

    @property
    def version(self) -> str:
        return self._version

    @property
    def is_admin(self) -> bool:
        if self._is_admin is None:
            response = self.GET("/api/v1/ActiveUser/Groups")
            self._is_admin = "ADMIN" in CaseAndSpaceInsensitiveSet(
                *[group["Name"] for group in response.json()["value"]])

        return self._is_admin

    @property
    def sandboxing_disabled(self):
        if self._sandboxing_disabled is None:
            value = self.GET("/api/v1/ActiveConfiguration/Administration/DisableSandboxing/$value")
            self._sandboxing_disabled = value

        return self._sandboxing_disabled

    @property
    def session_id(self) -> str:
        return self._s.cookies["TM1SessionId"]

    @staticmethod
    def translate_to_boolean(value) -> bool:
        """ Takes a boolean or string (eg. true, True, FALSE, etc.) value and returns (boolean) True or False
        :param value: True, 'true', 'false' or 'False' ...
        :return:
        """
        if isinstance(value, bool) or isinstance(value, int):
            return bool(value)
        elif isinstance(value, str):
            return value.replace(" ", "").lower() == 'true'
        else:
            raise ValueError("Invalid argument: '" + value + "'. Must be to be of type 'bool' or 'str'")

    @staticmethod
    def b64_decode_password(encrypted_password: str) -> str:
        """ b64 decoding
        :param encrypted_password: encrypted password with b64
        :return: password in plain text
        """
        return b64decode(encrypted_password).decode("UTF-8")

    @staticmethod
    def verify_response(response: Response):
        """ check if Status Code is OK
        :Parameters:
            `response`: String
                the response that is returned from a method call
        :Exceptions:
            TM1pyException, raises TM1pyException when Code is not 200, 204 etc.
        """
        if not response.ok:
            raise TM1pyRestException(response.text,
                                     status_code=response.status_code,
                                     reason=response.reason,
                                     headers=response.headers)

    @staticmethod
    def _build_authorization_token(user: str, password: str, namespace: str = None, gateway: str = None,
                                   cam_passport: str = None, verify: bool = False) -> str:
        """ Build the Authorization Header for CAM and Native Security
        """
        if cam_passport:
            return 'CAMPassport ' + cam_passport
        elif namespace:
            return RestService._build_authorization_token_cam(user, password, namespace, gateway, verify)
        else:
            return RestService._build_authorization_token_basic(user, password)

    @staticmethod
    def _build_authorization_token_cam(user: str = None, password: str = None, namespace: str = None,
                                       gateway: str = None, verify: bool = False) -> str:
        if gateway:
            try:
                HttpNegotiateAuth
            except NameError:
                raise RuntimeError(
                    "SSO failed due to missing dependency requests_negotiate_sspi.HttpNegotiateAuth. "
                    "SSO only supported for Windows")
            response = requests.get(gateway, auth=HttpNegotiateAuth(), verify=verify,
                                    params={"CAMNamespace": namespace})
            if not response.status_code == 200:
                raise RuntimeError(
                    "Failed to authenticate through CAM. Expected status_code 200, received status_code: "
                    + str(response.status_code))
            elif 'cam_passport' not in response.cookies:
                raise RuntimeError(
                    "Failed to authenticate through CAM. HTTP response does not contain 'cam_passport' cookie")
            else:
                return 'CAMPassport ' + response.cookies['cam_passport']
        else:
            return 'CAMNamespace ' + b64encode(str.encode("{}:{}:{}".format(user, password, namespace))).decode("ascii")

    @staticmethod
    def _build_authorization_token_basic(user: str, password: str) -> str:
        return 'Basic ' + b64encode(str.encode("{}:{}".format(user, password))).decode("ascii")

    @staticmethod
    def disable_http_warnings():
        # disable HTTP verification warnings from requests library
        requests.packages.urllib3.disable_warnings()

    def get_http_header(self, key: str) -> str:
        return self._headers[key]

    def add_http_header(self, key: str, value: str):
        self._headers[key] = value

    def remove_http_header(self, key: str):
        if key in self._headers:
            self._headers.pop(key)

    def add_compact_json_header(self) -> str:
        original_header = self.get_http_header('Accept')
        parts = original_header.split(';')

        # Point of insertion is important. Needs to come after application/json
        parts.insert(1, 'tm1.compact=v0')
        modified_header = ";".join(parts)
        self.add_http_header('Accept', modified_header)

        return original_header

    def retrieve_async_response(self, async_id: str, **kwargs) -> Response:
        url = self._base_url + f"/api/v1/_async('{async_id}')"
        return self._s.get(url, **kwargs)

    def cancel_async_operation(self, async_id: str, **kwargs):
        url = self._base_url + f"/api/v1/_async('{async_id}')"
        response = self._s.delete(url, **kwargs)
        self.verify_response(response)

    def cancel_running_operation(self):
        monitoring_service = self.get_monitoring_service()
        threads = monitoring_service.get_active_session_threads(exclude_idle=True)

        # if more than one thread is running in session, operation can not be identified unambiguously
        if not len(threads) == 1:
            return

        monitoring_service.cancel_thread(threads[0]['ID'])

    def get_monitoring_service(self):
        from TM1py.Services import MonitoringService
        return MonitoringService(self)

    @staticmethod
    def urllib3_response_from_bytes(data: bytes) -> HTTPResponse:
        sock = BytesIOSocket(data)

        response = HTTPResponse(sock)
        response.begin()

        return urllib3.HTTPResponse.from_httplib(response)

    @staticmethod
    def build_response_from_raw_bytes(data: bytes) -> Response:
        urllib_response = RestService.urllib3_response_from_bytes(data)

        adapter = HTTPAdapter()
        requests_response = adapter.build_response(requests.PreparedRequest(), urllib_response)
        # actual content of response needs to be set explicitly
        requests_response._content = urllib_response.data

        return requests_response

    @staticmethod
    def wait_time_generator(timeout: int):
        yield 0.1
        yield 0.3
        yield 0.6
        if timeout:
            for _ in range(1, int(timeout)):
                yield 1
        else:
            while True:
                yield 1

    def _determine_ssl_based_on_base_url(self) -> bool:
        if self._base_url.startswith("https"):
            return True
        elif self._base_url.startswith("http"):
            return False
        else:
            raise ValueError(f"Invalid base_url: '{self._base_url}'")


class BytesIOSocket:
    """ used in urllib3_response_from_bytes method to construct urllib3 response from raw bytes

    """

    def __init__(self, content: bytes):
        self.handle = BytesIO(content)

    def makefile(self, mode) -> BytesIO:
        return self.handle



# -*- coding: utf-8 -*-

import json
from enum import Enum
from typing import Dict, Union

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import case_and_space_insensitive_equals


class ElementAttribute(TM1Object):
    """ Abstraction of TM1 Element Attributes
    
    """

    class Types(Enum):
        NUMERIC = 1
        STRING = 2
        ALIAS = 3

        def __str__(self):
            return self.name.capitalize()

        @classmethod
        def _missing_(cls, value: str):
            for member in cls:
                if member.name.lower() == value.replace(" ", "").lower():
                    return member
            # default
            raise ValueError(f"Invalid attribute type: '{value}'")

    def __init__(self, name: str, attribute_type: Union[Types, str]):
        self.name = name
        self.attribute_type = attribute_type

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def attribute_type(self) -> str:
        return str(self._attribute_type)

    @attribute_type.setter
    def attribute_type(self, value: Union[Types, str]):
        self._attribute_type = ElementAttribute.Types(value)

    @property
    def body_as_dict(self) -> Dict:
        return {"Name": self._name, "Type": self.attribute_type}

    @property
    def body(self) -> str:
        return json.dumps(self.body_as_dict, ensure_ascii=False)

    @classmethod
    def from_json(cls, element_attribute_as_json: str) -> 'ElementAttribute':
        return cls.from_dict(json.loads(element_attribute_as_json))

    @classmethod
    def from_dict(cls, element_attribute_as_dict: Dict) -> 'ElementAttribute':
        return cls(name=element_attribute_as_dict['Name'],
                   attribute_type=element_attribute_as_dict['Type'])

    def __eq__(self, other: Union[str, 'ElementAttribute']):
        if isinstance(other, str):
            return case_and_space_insensitive_equals(self.name, other)
        elif isinstance(other, ElementAttribute):
            return case_and_space_insensitive_equals(self.name, other.name)
        else:
            raise ValueError("Argument: 'other' must be of type str or ElementAttribute")

    def __hash__(self):
        return super().__hash__()



# -*- coding: utf-8 -*-
from typing import List

from TM1py.Objects.GitCommit import GitCommit


class GitPlan:
    """ Base GitPlan abstraction
    """

    def __init__(self, plan_id: str, branch: str, force: bool):
        """ Initialize GitPlan object
        :param plan_id: id of the Plan
        :param branch: current branch
        :param force: force git context reset
        """
        self._plan_id = plan_id
        self._branch = branch
        self._force = force

    @property
    def plan_id(self) -> str:
        return self._plan_id

    @property
    def branch(self) -> str:
        return self._branch

    @property
    def force(self) -> bool:
        return self._force


class GitPushPlan(GitPlan):
    """ GitPushPlan abstraction based on GitPlan
    """

    def __init__(self, plan_id: str, branch: str, force: bool, new_branch: str, new_commit: GitCommit,
                 parent_commit: GitCommit, source_files: List[str]):
        """ Initialize GitPushPlan object
        :param plan_id: id of the PushPlan
        :param branch: current branch to base the pushplan on
        :param force: force git context reset
        :param new_branch: the new branch that will be pushed to
        :param new_commit: GitCommit of the new commit
        :param parent_commit: The current commit in the branch
        :param source_files: list of included files in the push
        """
        self._new_branch = new_branch
        self._new_commit = new_commit
        self._parent_commit = parent_commit
        self._source_files = source_files

        super().__init__(plan_id=plan_id, branch=branch, force=force)

    @property
    def new_branch(self) -> str:
        return self._new_branch

    @property
    def new_commit(self) -> GitCommit:
        return self._new_commit

    @property
    def parent_commit(self) -> GitCommit:
        return self._parent_commit

    @property
    def source_files(self) -> List[str]:
        return self._source_files


class GitPullPlan(GitPlan):
    """ GitPushPlan abstraction based on GitPlan
    """

    def __init__(self, plan_id: str, branch: str, force: bool, commit: GitCommit, operations: List[str]):
        """ Initialize GitPushPlan object
        :param plan_id: id of the PullPlan
        :param branch: current branch to base the pullplan on
        :param force: force git context reset
        :param commit: GitCommit of the commit to pull
        :param operations: list of changes made upon pulling
        """
        self._commit = commit
        self._operations = operations

        super().__init__(plan_id=plan_id, branch=branch, force=force)

    @property
    def commit(self) -> GitCommit:
        return self._commit

    @property
    def operations(self) -> List[str]:
        return self._operations



# -*- coding: utf-8 -*-
import json
import random
from typing import List, Iterable, Dict

from requests import Response

from TM1py.Objects.Cube import Cube
from TM1py.Services.CellService import CellService
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Services.ViewService import ViewService
from TM1py.Utils import format_url, require_version, require_admin, case_and_space_insensitive_equals


class CubeService(ObjectService):
    """ Service to handle Object Updates for TM1 Cubes

    """

    def __init__(self, rest: RestService):
        # to avoid Circular dependency of modules
        from TM1py.Services.AnnotationService import AnnotationService
        super().__init__(rest)
        self.cells = CellService(rest)
        self.views = ViewService(rest)
        self.annotations = AnnotationService(rest)

    def create(self, cube: Cube, **kwargs) -> Response:
        """ create new cube on TM1 Server

        :param cube: instance of TM1py.Cube
        :return: response
        """
        url = "/api/v1/Cubes"
        return self._rest.POST(url=url, data=cube.body, **kwargs)

    def get(self, cube_name: str, **kwargs) -> Cube:
        """ get cube from TM1 Server

        :param cube_name:
        :return: instance of TM1py.Cube
        """
        url = format_url("/api/v1/Cubes('{}')?$expand=Dimensions($select=Name)", cube_name)
        response = self._rest.GET(url=url, **kwargs)
        cube = Cube.from_json(response.text)
        # cater for potential EnableSandboxDimension=T setup
        if case_and_space_insensitive_equals(cube.dimensions[0], "Sandboxes"):
            cube.dimensions = cube.dimensions[1:]
        return cube

    def get_last_data_update(self, cube_name: str, **kwargs) -> str:
        url = format_url("/api/v1/Cubes('{}')/LastDataUpdate/$value", cube_name)
        response = self._rest.GET(url=url, **kwargs)
        return response.text

    def get_all(self, **kwargs) -> List[Cube]:
        """ get all cubes from TM1 Server as TM1py.Cube instances

        :return: List of TM1py.Cube instances
        """
        url = "/api/v1/Cubes?$expand=Dimensions($select=Name)"
        response = self._rest.GET(url, **kwargs)
        cubes = [Cube.from_dict(cube_as_dict=cube) for cube in response.json()['value']]
        return cubes

    def get_model_cubes(self, **kwargs) -> List[Cube]:
        """ Get all Cubes without } prefix from TM1 Server as TM1py.Cube instances

        :return: List of TM1py.Cube instances
        """
        url = "/api/v1/ModelCubes()?$expand=Dimensions($select=Name)"
        response = self._rest.GET(url, **kwargs)
        cubes = [Cube.from_dict(cube_as_dict=cube) for cube in response.json()['value']]
        return cubes

    def get_control_cubes(self, **kwargs) -> List[Cube]:
        """ Get all Cubes with } prefix from TM1 Server as TM1py.Cube instances

        :return: List of TM1py.Cube instances
        """
        url = "/api/v1/ControlCubes()?$expand=Dimensions($select=Name)"
        response = self._rest.GET(url, **kwargs)
        cubes = [Cube.from_dict(cube_as_dict=cube) for cube in response.json()['value']]
        return cubes

    def get_number_of_cubes(self, skip_control_cubes: bool = False, **kwargs) -> int:
        """ Ask TM1 Server for count of cubes

        :skip_control_cubes: bool, True will exclude control cubes from count
        :return: int, count
        """ 
        if skip_control_cubes:
            response = self._rest.GET(url=format_url("/api/v1/ModelCubes()?$select=Name&$top=0&$count"), **kwargs)
            return int(response.json()['@odata.count'])

        return int(self._rest.GET(url=format_url("/api/v1/Cubes/$count"), **kwargs).text)

    def get_measure_dimension(self, cube_name: str, **kwargs) -> str:
        url = format_url(
            "/api/v1/Cubes('{}')/Dimensions?$select=Name",
            cube_name)
        response = self._rest.GET(url, **kwargs)
        return response.json()['value'][-1]['Name']

    def update(self, cube: Cube, **kwargs) -> Response:
        """ Update existing cube on TM1 Server

        :param cube: instance of TM1py.Cube
        :return: response
        """
        url = format_url("/api/v1/Cubes('{}')", cube.name)
        return self._rest.PATCH(url, cube.body, **kwargs)

    def update_or_create(self, cube: Cube, **kwargs) -> Response:
        """ update if exists else create

        :param cube:
        :return:
        """
        if self.exists(cube_name=cube.name, **kwargs):
            return self.update(cube=cube, **kwargs)

        return self.create(cube=cube, **kwargs)

    def check_rules(self, cube_name: str, **kwargs) -> Response:
        """ Check rules syntax for existing cube on TM1 Server

        :param cube_name: name of a cube
        :return: response
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.CheckRules", cube_name)

        response = self._rest.POST(url, **kwargs)
        errors = response.json()["value"]
        return errors

    @require_admin
    def delete(self, cube_name: str, **kwargs) -> Response:
        """ Delete a cube in TM1

        :param cube_name:
        :return: response
        """
        url = format_url("/api/v1/Cubes('{}')", cube_name)
        return self._rest.DELETE(url, **kwargs)

    def exists(self, cube_name: str, **kwargs) -> bool:
        """ Check if a cube exists. Return boolean.

        :param cube_name: 
        :return: Boolean 
        """
        url = format_url("/api/v1/Cubes('{}')", cube_name)
        return self._exists(url, **kwargs)

    def get_all_names(self, skip_control_cubes: bool = False, **kwargs) -> List[str]:
        """ Ask TM1 Server for list of all cube names
        
        :skip_control_cubes: bool, True will exclude control cubes from list
        :return: List of Strings
        """
        url = format_url(
                "/api/v1/{}?$select=Name",
                'ModelCubes()' if skip_control_cubes else 'Cubes'
        )

        response = self._rest.GET(url, **kwargs)
        cubes = list(entry['Name'] for entry in response.json()['value'])
        return cubes

    def get_all_names_with_rules(self, skip_control_cubes: bool = False, **kwargs) -> List[str]:
        """ Ask TM1 Server for list of all cube names that have rules
        
        :skip_control_cubes: bool, True will exclude control cubes from list
        :return: List of Strings
        """
        url = format_url(
                "/api/v1/{}?$select=Name,Rules&$filter=Rules ne null",
                'ModelCubes()' if skip_control_cubes else 'Cubes'
        )

        response = self._rest.GET(url, **kwargs)
        cubes = list(cube['Name'] for cube in response.json()['value'])
        return cubes

    def get_all_names_without_rules(self, skip_control_cubes: bool = False, **kwargs) -> List[str]:
        """ Ask TM1 Server for list of all cube names that do not have rules
        :skip_control_cubes: bool, True will exclude control cubes from list
        :return: List of Strings
        """

        url = format_url(
                "/api/v1/{}?$select=Name,Rules&$filter=Rules eq null",
                'ModelCubes()' if skip_control_cubes else 'Cubes'
        )

        response = self._rest.GET(url, **kwargs)
        cubes = list(cube['Name'] for cube in response.json()['value'])
        return cubes

    def get_dimension_names(self, cube_name: str, skip_sandbox_dimension: bool = True, **kwargs) -> List[str]:
        """ get name of the dimensions of a cube in their correct order

        :param cube_name:
        :param skip_sandbox_dimension:
        :return:  List : [dim1, dim2, dim3, etc.]
        """
        url = format_url("/api/v1/Cubes('{}')/Dimensions?$select=Name", cube_name)
        response = self._rest.GET(url, **kwargs)
        dimension_names = [element['Name'] for element in response.json()['value']]
        if skip_sandbox_dimension and dimension_names[0] == CellService.SANDBOX_DIMENSION:
            return dimension_names[1:]
        return dimension_names

    def search_for_dimension(self, dimension_name: str, skip_control_cubes: bool = False,
                             **kwargs) -> List[str]:
        """ Ask TM1 Server for list of cube names that contain specific dimension

        :param dimension_name: string, valid dimension name (case insensitive)
        :param skip_control_cubes: bool, True will exclude control cubes from result
        """
        url = format_url(
            "/api/v1/{}?$select=Name&$filter=Dimensions/any(d: replace(tolower(d/Name), ' ', '') eq '{}')",
            'ModelCubes()' if skip_control_cubes else 'Cubes',
            dimension_name.lower().replace(' ', '')
        )
        response = self._rest.GET(url, **kwargs)
        cubes = list(entry['Name'] for entry in response.json()['value'])
        return cubes

    def search_for_dimension_substring(self, substring: str, skip_control_cubes: bool = False,
                                       **kwargs) -> Dict[str, List[str]]:
        """ Ask TM1 Server for a dictinary of cube names with the dimension whose name contains the substring

        :param substring: string to search for in dim name
        :param skip_control_cubes: bool, True will exclude control cubes from result
        """
        substring = substring.lower().replace(' ', '')

        url = format_url(
            "/api/v1/{}?$select=Name&$filter=Dimensions/any(d: contains(replace(tolower(d/Name), ' ', ''),'{}'))" +
            "&$expand=Dimensions($select=Name;$filter=contains(replace(tolower(Name), ' ', ''), '{}'))",
            'ModelCubes()' if skip_control_cubes else 'Cubes',
            substring,
            substring)

        response = self._rest.GET(url, **kwargs)
        cube_dict = {entry['Name']: [dim['Name'] for dim in entry['Dimensions']] for entry in response.json()['value']}
        return cube_dict
    
    def search_for_rule_substring(self, substring: str, skip_control_cubes: bool = False, case_insensitive=True,
                                  space_insensitive=True, **kwargs) -> List[Cube]:
        """ get all cubes from TM1 Server as TM1py.Cube instances where rules for given cube contain specified substring

        :param substring: string to search for in rules
        :param skip_control_cubes: bool, True will exclude control cubes from result
        :param case_insensitive: case agnostic search
        :param space_insensitive: space agnostic search
        :return: List of TM1py.Cube instances
        """
        substring = substring.lower().replace(' ', '')

        url_filter = "Rules ne null and contains("
        if case_insensitive and space_insensitive:
            url_filter += format_url("tolower(replace(Rules, ' ', '')),'{}')", substring)
        elif case_insensitive:
            url_filter += format_url("tolower(Rules),'{}')", substring)
        elif space_insensitive:
            url_filter += format_url("replace(Rules, ' ', ''),'{}')", substring)
        else:
            url_filter += format_url("Rules,'{}')", substring)

        url = f"/api/v1/{'ModelCubes()' if skip_control_cubes else 'Cubes'}?$filter={url_filter}" \
              f"&$expand=Dimensions($select=Name)"

        response = self._rest.GET(url, **kwargs)
        cubes = [Cube.from_dict(cube_as_dict=cube) for cube in response.json()['value']]
        return cubes   

    @require_version(version="11.4")
    def get_storage_dimension_order(self, cube_name: str, **kwargs) -> List[str]:
        """ Get the storage dimension order of a cube

        :param cube_name:
        :return: List of dimension names
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.DimensionsStorageOrder()?$select=Name", cube_name)
        response = self._rest.GET(url, **kwargs)
        return [dimension["Name"] for dimension in response.json()["value"]]

    @require_admin
    @require_version(version="11.4")
    def update_storage_dimension_order(self, cube_name: str, dimension_names: Iterable[str]) -> float:
        """ Update the storage dimension order of a cube

        :param cube_name:
        :param dimension_names:
        :return:  Float: -23.076489699337078 (percent change in memory usage)
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.ReorderDimensions", cube_name)
        payload = dict()
        payload['Dimensions@odata.bind'] = [format_url("Dimensions('{}')", dimension)
                                            for dimension
                                            in dimension_names]
        response = self._rest.POST(url=url, data=json.dumps(payload))
        return response.json()["value"]

    @require_admin
    @require_version(version="11.6")
    def load(self, cube_name: str, **kwargs) -> Response:
        """ Load the cube into memory on the server

        :param cube_name:
        :return:
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.Load", cube_name)
        return self._rest.POST(url=url, **kwargs)

    @require_admin
    @require_version(version="11.6")
    def unload(self, cube_name: str, **kwargs) -> Response:
        """ Unload the cube from memory

        :param cube_name:
        :return:
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.Unload", cube_name)
        return self._rest.POST(url=url, **kwargs)

    def lock(self, cube_name: str, **kwargs) -> Response:
        """ Locks the cube to prevent any users from modifying it

        :param cube_name:
        :return:
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.Lock", cube_name)
        return self._rest.POST(url=url, **kwargs)

    def unlock(self, cube_name: str, **kwargs) -> Response:
        """ Unlocks the cube to allow modifications

        :param cube_name:
        :return:
        """
        url = format_url("/api/v1/Cubes('{}')/tm1.Unlock", cube_name)
        return self._rest.POST(url=url, **kwargs)

    @require_admin
    def cube_save_data(self, cube_name: str, **kwargs) -> Response:
        """ Serializes a cube by saving data updates

        :param cube_name:
        :return: Response
        """
        from TM1py.Services import ProcessService
        ti = "CubeSaveData('{0}');".format(cube_name)
        process_service = ProcessService(self._rest)
        return process_service.execute_ti_code(ti, **kwargs)

    def get_random_intersection(self, cube_name: str, unique_names: bool = False) -> List[str]:
        """ Get a random Intersection in a cube
        used mostly for regression testing.
        Not optimized, in terms of performance. Function Loads ALL elements for EACH dim...

        :param cube_name: 
        :param unique_names: unique names instead of plain element names 
        :return: List of elements
        """
        from TM1py.Services import DimensionService
        dimension_service = DimensionService(self._rest)
        dimensions = self.get_dimension_names(cube_name)
        elements = []
        for dimension in dimensions:
            hierarchy = dimension_service.get(dimension).default_hierarchy
            element = random.choice(list((hierarchy.elements.keys())))
            if unique_names:
                element = '[{}].[{}]'.format(dimension, element)
            elements.append(element)
        return elements



# -*- coding: utf-8 -*-
from typing import Union

from TM1py.Objects.TM1Object import TM1Object


class ChoreFrequency(TM1Object):
    """ Utility class to handle time representation fore Chore Frequency
    
    """

    def __init__(self, days: Union[str, int], hours: Union[str, int], minutes: Union[str, int],
                 seconds: Union[str, int]):
        self._days = str(days).zfill(2)
        self._hours = str(hours).zfill(2)
        self._minutes = str(minutes).zfill(2)
        self._seconds = str(seconds).zfill(2)

    @property
    def days(self) -> str:
        return self._days

    @property
    def hours(self) -> str:
        return self._hours

    @property
    def minutes(self) -> str:
        return self._minutes

    @property
    def seconds(self) -> str:
        return self._seconds

    @days.setter
    def days(self, value: Union[str, int]):
        self._days = str(value).zfill(2)

    @hours.setter
    def hours(self, value: Union[str, int]):
        self._hours = str(value).zfill(2)

    @minutes.setter
    def minutes(self, value: Union[str, int]):
        self._minutes = str(value).zfill(2)

    @seconds.setter
    def seconds(self, value: Union[str, int]):
        self._seconds = str(value).zfill(2)

    @classmethod
    def from_string(cls, frequency_string: str) -> 'ChoreFrequency':
        pos_dt = frequency_string.find('DT', 1)
        pos_h = frequency_string.find('H', pos_dt)
        pos_m = frequency_string.find('M', pos_h)
        pos_s = len(frequency_string) - 1
        return cls(days=frequency_string[1:pos_dt],
                   hours=frequency_string[pos_dt + 2:pos_h],
                   minutes=frequency_string[pos_h + 1:pos_m],
                   seconds=frequency_string[pos_m + 1:pos_s])

    @property
    def frequency_string(self) -> str:
        return "P{}DT{}H{}M{}S".format(self._days, self._hours, self._minutes, self._seconds)

    def __str__(self) -> str:
        return self.frequency_string



# -*- coding: utf-8 -*-
from typing import Dict


class Server:
    """ Abstraction of the TM1 Server

        :Notes:
            contains the information you get from http://localhost:5895/api/v1/Servers
            no methods so far
    """
    def __init__(self, server_as_dict: Dict):
        self.name = server_as_dict['Name']
        self.ip_address = server_as_dict['IPAddress']
        self.ip_v6_address = server_as_dict['IPv6Address']
        self.port_number = server_as_dict['PortNumber']
        self.client_message_port_number = server_as_dict['ClientMessagePortNumber']
        self.http_port_number = server_as_dict['HTTPPortNumber']
        self.using_ssl = server_as_dict['UsingSSL']
        self.accepting_clients = server_as_dict['AcceptingClients']
        self.self_registered = server_as_dict['SelfRegistered'] 
        self.host = server_as_dict['Host'] 
        self.is_local = server_as_dict['IsLocal'] 
        self.ssl_certificate_id = server_as_dict['SSLCertificateID'] 
        self.ssl_certificate_authority = server_as_dict['SSLCertificateAuthority'] 
        self.ssl_certificate_revocation_list = server_as_dict['SSLCertificateRevocationList'] 
        self.client_export_ssl_server_keyid = server_as_dict['ClientExportSSLSvrKeyID'] 
        self.client_export_ssl_server_cert = server_as_dict['ClientExportSSLSvrCert'] 
        self.last_updated = server_as_dict['LastUpdated']